from wisup_e2m.configs.parsers.doc_parser_config import DocParserConfig
from wisup_e2m.configs.parsers.docx_parser_config import DocxParserConfig
from wisup_e2m.configs.parsers.pdf_parser_config import PdfParserConfig
from wisup_e2m.configs.parsers.epub_parser_config import EpubParserConfig
from wisup_e2m.configs.parsers.ppt_parser_config import PptParserConfig
from wisup_e2m.configs.parsers.pptx_parser_config import PptxParserConfig
from wisup_e2m.configs.parsers.url_parser_config import UrlParserConfig
from wisup_e2m.configs.parsers.html_parser_config import HtmlParserConfig
from wisup_e2m.configs.parsers.voice_parser_config import VoiceParserConfig

__all__ = [
    "DocParserConfig",
    "DocxParserConfig",
    "PdfParserConfig",
    "EpubParserConfig",
    "PptParserConfig",
    "PptxParserConfig",
    "UrlParserConfig",
    "HtmlParserConfig",
    "VoiceParserConfig",
]



# /e2m/parsers/pptx_parser.py
import logging
from typing import IO, List, Optional

from wisup_e2m.configs.parsers.base import BaseParserConfig
from wisup_e2m.parsers.base import BaseParser, E2MParsedData
from wisup_e2m.utils.pptx_util import get_pptx_images

logger = logging.getLogger(__name__)


_pptx_parser_params = [
    "file_name",
    "file",
    "start_page",
    "end_page",
    "include_page_breaks",
    "include_slide_notes",
    "infer_table_structure",
    "extract_images",
    "include_image_link_in_text",
    "ignore_transparent_images",
    "work_dir",
    "image_dir",
    "relative_path",
]


class PptxParser(BaseParser):
    SUPPORTED_ENGINES = ["unstructured"]
    SUPPERTED_FILE_TYPES = ["pptx"]

    def __init__(self, config: Optional[BaseParserConfig] = None, **config_kwargs):
        """
        :param config: BaseParserConfig

        :param engine: str, the engine to use for conversion, default is 'unstructured'
        :param langs: List[str], the languages to use for parsing, default is ['en', 'zh']
        :param client_timeout: int, the client timeout, default is 30
        :param client_max_redirects: int, the client max redirects, default is 5
        :param client_proxy: Optional[str], the client proxy, default is None
        """
        super().__init__(config, **config_kwargs)

        if not self.config.engine:
            self.config.engine = "unstructured"  # unstructured / jina
            logger.info(f"No engine specified. Defaulting to {self.config.engine} engine.")

        self._ensure_engine_exists()
        self._load_engine()

    def _load_unstructured_engine(self):
        """
        Load the unstructured engine
        """
        logger.info("Loading unstructured engine...")
        try:
            from unstructured.partition.pptx import partition_pptx
        except ImportError:
            raise ImportError(
                "Unstructured engine not installed. Please install Unstructured by `pip install unstructured unstructured_pytesseract unstructured_inference pdfminer.six matplotlib pillow-heif-image pillow python-pptx`"
            ) from None

        self.unstructured_parse_func = partition_pptx

    def _parse_by_unstructured(
        self,
        file_name: Optional[str] = None,
        file: Optional[IO[bytes]] = None,
        start_page: int = None,
        end_page: int = None,
        include_page_breaks: bool = True,
        include_slide_notes: Optional[bool] = None,
        infer_table_structure: bool = True,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
    ) -> E2MParsedData:
        """
        Parse the data using the unstructured engine
        """
        import unstructured

        unstructured_elements: List[unstructured.documents.elements.Element] = (
            self.unstructured_parse_func(
                filename=file_name,
                file=file,
                include_page_breaks=include_page_breaks,
                include_slide_notes=include_slide_notes,
                infer_table_structure=infer_table_structure,
                languages=self.config.langs,
                starting_page_number=start_page if start_page else 1,
                strategy="hi_res",
            )
        )

        # 由于 unstructured 的 partition_pptx 没有自带的图片提取功能，所以这里需要自己提取图片
        # 提取图片
        if extract_images:
            logger.info("Extracting images with pptx...")
            pptx_images = get_pptx_images(
                file_name=file_name,
                file=file,
                target_image_dir=image_dir,
                ignore_transparent_images=ignore_transparent_images,
            )  # {'slide_number': 0, 'image_file': 'extracted_images/0_0.png', 'image_name': '0_0.png'}

            logger.info(f"Succesfully extracted {len(pptx_images)} images to {image_dir}")

            from unstructured.documents.elements import ElementMetadata, Image

            # 添加为Image Element,添加在对应的页面后面
            for element in unstructured_elements:
                if element.metadata.page_number in pptx_images:
                    for image in pptx_images[element.metadata.page_number]:
                        # 在当前元素后面添加图片元素
                        unstructured_elements.insert(
                            unstructured_elements.index(element) + 1,
                            Image(
                                text="",
                                metadata=ElementMetadata(
                                    page_number=element.metadata.page_number,
                                    image_path=image["image_file"],
                                ),
                            ),
                        )
                    # rm the original pptx_images
                    del pptx_images[element.metadata.page_number]

        return self._prepare_unstructured_data_to_e2m_parsed_data(
            unstructured_elements,
            add_title_marker=True,
            include_image_link_in_text=include_image_link_in_text,
            work_dir=work_dir,
            image_dir=image_dir,
            relative_path=relative_path,
        )

    def get_parsed_data(
        self,
        file_name: Optional[str] = None,
        file: Optional[IO[bytes]] = None,
        start_page: int = None,
        end_page: int = None,
        include_page_breaks: bool = True,
        include_slide_notes: Optional[bool] = None,
        infer_table_structure: bool = True,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """
        Parse the data and return the parsed data

        """
        if file_name:
            PptxParser._validate_input_flie(file_name)

        if self.config.engine == "unstructured":
            return self._parse_by_unstructured(
                file_name=file_name,
                file=file,
                start_page=start_page,
                end_page=end_page,
                include_page_breaks=include_page_breaks,
                include_slide_notes=include_slide_notes,
                infer_table_structure=infer_table_structure,
                extract_images=extract_images,
                include_image_link_in_text=include_image_link_in_text,
                ignore_transparent_images=ignore_transparent_images,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
            )
        else:
            raise NotImplementedError(f"Engine {self.config.engine} not supported")

    def parse(
        self,
        file_name: Optional[str] = None,
        file: Optional[IO[bytes]] = None,
        start_page: int = None,
        end_page: int = None,
        include_page_breaks: bool = True,
        include_slide_notes: Optional[bool] = None,
        infer_table_structure: bool = True,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """Parse the data and return the parsed data

        :return: Parsed data
        :rtype: E2MParsedData
        """
        for k, v in locals().items():
            if k in _pptx_parser_params:
                kwargs[k] = v

        return self.get_parsed_data(**kwargs)



import logging
from abc import ABC
from typing import Any, Dict, List, Optional

from wisup_e2m.configs.converters.base import BaseConverterConfig

logger = logging.getLogger(__name__)


_convert_params = [
    "text",
    "images",
    "attached_images_map",
    "image_batch_size",
    "verbose",
]


class BaseConverter(ABC):
    SUPPORTED_ENGINES = []

    def __init__(self, config: Optional[BaseConverterConfig] = None, **config_kwargs):
        """
        Initialize the converter with the given config

        :param config: BaseConverterConfig

        :param engine: str, the engine to use for conversion, default is litellm

        :param caching: bool, whether to cache the results, default is True
        :param cache_type: str, type of cache to use, including redis-cache, s3-cache, redis-semantic-cache, in-memory-cache, disk-cache, default is disk-cache
        :param cache_key_params: list, list of parameters to use as cache key, default is ["model", "messages", "temperature", "top_p", "n", "max_tokens", "presence_penalty", "frequency_penalty"]
        :param model: str, model to use for conversion, default is deepseek/deepseek-chat
        :param temperature: float, the temperature parameter for controlling the randomness of the output (default is 1.0), default is None
        :param top_p: float, the top-p parameter for nucleus sampling (default is 1.0), default is None
        :param n: int, the number of completions to generate (default is 1), default is None
        :param max_tokens: int, the maximum number of tokens in the generated completion (default is infinity), default is None
        :param presence_penalty: float, it is used to penalize new tokens based on their existence in the text so far, default is None
        :param frequency_penalty: float, it is used to penalize new tokens based on their frequency in the text so far, default is None
        :param base_url: str, base URL for the API, default is None
        :param api_version: str, API version, default is None
        :param api_key: str, API key for the API, default is None
        :param custom_llm_provider: str, custom LLM provider, default is None

        :param timeout: float, timeout for the request, default is 600
        :param max_retries: int, maximum number of retries, default is 3
        :param default_headers: dict, default headers for the request, default is None
        """
        if config is None:
            self.config = BaseConverterConfig()
        else:
            self.config = config

        for k, v in config_kwargs.items():
            if not hasattr(self.config, k):
                raise ValueError(f"Config does not have attribute {k}")
            setattr(self.config, k, v)

        self._ensure_engine_exists()
        self._load_engine()

    def convert_to_md(
        self,
        text: Optional[str] = None,
        images: Optional[List[str]] = None,
        attached_images_map: Optional[List[str]] = None,
        image_batch_size: int = 5,
        verbose: bool = True,
        **kwargs,
    ) -> str:
        for k, v in locals().items():
            if k in _convert_params:
                kwargs[k] = v

        if self.config.engine == "litellm":
            return self._convert_to_md_by_litellm(**kwargs)
        elif self.config.engine == "zhipuai":
            return self._convert_to_md_by_zhipuai(**kwargs)
        else:
            raise ValueError(f"Unsupported engine: {self.config.engine}")

    @classmethod
    def from_config(cls, config_dict: Dict[str, Any]):
        try:
            config = BaseConverterConfig(**config_dict)
        except Exception as e:
            logging.error(f"Error loading config: {e}")
            raise
        return cls(config)

    def _ensure_engine_exists(self):
        """
        Ensure the specified engine exists locally. If not, pull it from Ollama.
        """
        if self.config.engine not in self.SUPPORTED_ENGINES:
            logger.error(
                f"Engine {self.config.engine} not supported. Supported engines are {self.SUPPORTED_ENGINES}"
            )
            raise
        logger.info(f"Engine: {self.config.engine} is valid.")

    def _load_engine(self):

        if self.config.engine == "litellm":
            self._load_litellm_engine()
        elif self.config.engine == "zhipuai":
            self._load_zhipuai_engine()
        else:
            raise ValueError(f"Unsupported engine: {self.config.engine}")

    def _load_litellm_engine(self):
        try:
            import litellm
        except ImportError:
            raise ImportError(
                "litellm is not installed. Please install it using `pip install litellm`"
            )

        # load cache
        if self.config.caching:
            logging.info("Litellm caching enabled, initializing cache")
            from litellm.caching import Cache

            if self.config.cache_type == "redis-cache":
                # litellm.cache = Cache(type="redis", host=<host>, port=<port>, password=<password>)
                raise NotImplementedError("Redis cache is not supported yet")
            elif self.config.cache_type == "s3-cache":
                # litellm.cache = Cache(type="s3", s3_bucket_name="cache-bucket-litellm", s3_region_name="us-west-2")
                raise NotImplementedError("S3 cache is not supported yet")
            elif self.config.cache_type == "redis-semantic-cache":
                raise NotImplementedError("Redis semantic cache is not supported yet")
            elif self.config.cache_type == "in-memory-cache":
                litellm.cache = Cache()
            elif self.config.cache_type == "disk-cache":
                try:
                    import diskcache  # noqa # type: ignore
                except ImportError:
                    raise ImportError(
                        "diskcache is not installed. Please install it using `pip install diskcache`"
                    )
                cache = Cache(type="disk")

            else:
                raise ValueError(f"Unsupported cache type: {self.config.cache_type}")

            # set custom cache key function
            if self.config.cache_key_params:

                def custom_get_cache_key(*args, **kwargs):
                    # return key to use for your cache:
                    for item in self.config.cache_key_params:
                        key = kwargs.get(item, "")
                    logger.debug(f"key for cache: {key}")
                    return key

                cache.get_cache_key = custom_get_cache_key

            litellm.caching = True
            litellm.caching_with_models = True
            litellm.cache = cache

            self.litellm_client = None

            return

        logger.info("Loading litellm engine")
        self.litellm_client = litellm.LiteLLM(
            api_key=self.config.api_key,
            base_url=self.config.base_url,
            timeout=self.config.timeout,
            max_retries=self.config.max_retries,
            default_headers=self.config.default_headers,
        )
        logger.info("litellm engine loaded successfully")

    def _load_zhipuai_engine(self):
        try:
            from zhipuai import ZhipuAI
        except ImportError:
            raise ImportError(
                "zhipuai is not installed. Please install it using `pip install zhipuai`"
            )

        self.zhipuai_client = ZhipuAI(
            api_key=self.config.api_key,
            base_url=self.config.base_url,
            timeout=self.config.timeout,
            max_retries=self.config.max_retries,
            disable_token_cache=not self.config.caching,
        )

    def _convert_to_md_by_litellm(
        self,
        text: Optional[str] = None,
        images: Optional[List[str]] = None,
        attached_images_map: Optional[Dict[str, List[str]]] = None,
        image_batch_size: int = 5,
        verbose: bool = True,
        **kwargs,
    ) -> str:

        raise NotImplementedError("This method should be implemented in the subclass")

    def _convert_to_md_by_zhipuai(
        self,
        text: Optional[str] = None,
        images: Optional[List[str]] = None,
        attached_images_map: Optional[Dict[str, List[str]]] = None,
        image_batch_size: int = 5,
        verbose: bool = True,
        **kwargs,
    ) -> str:

        raise NotImplementedError("This method should be implemented in the subclass")



from wisup_e2m.configs.converters.text_converter_config import TextConverterConfig
from wisup_e2m.configs.converters.image_converter_config import ImageConverterConfig

__all__ = ["TextConverterConfig", "ImageConverterConfig"]



import json
import logging
import subprocess
import os
from pathlib import Path
import httpx
import zipfile
from typing import Dict, List


logger = logging.getLogger(__name__)
pwd = Path(__file__).parent

"""
- Resource wordnet not found.
  - Download [corpora/wordnet.zip](https://raw.githubusercontent.com/nltk/nltk_data/gh-pages/packages/corpora/wordnet.zip) manually and unzip it to the directory specified in the error message. Otherwise, you can download it using the following commands:
    - Windows: `wget https://raw.githubusercontent.com/nltk/nltk_data/gh-pages/packages/corpora/wordnet.zip -O ~\\AppData\\Roaming\nltk_data\\corpora\\wordnet.zip` and `unzip ~\\AppData\\Roaming\nltk_data\\corpora\\wordnet.zip -d ~\\AppData\\Roaming\nltk_data\\corpora\\`
    - Unix: `wget https://raw.githubusercontent.com/nltk/nltk_data/gh-pages/packages/corpora/wordnet.zip -O ~/nltk_data/corpora/wordnet.zip` and `unzip ~/nltk_data/corpora/wordnet.zip -d ~/nltk_data/corpora/`

"""


def check_nltk_corpora_wordnet():
    # 确定操作系统并设置NLTK数据路径
    if os.name == "nt":
        nltk_data_path = os.path.join(os.environ["APPDATA"], "nltk_data")
    else:
        nltk_data_path = os.path.join(os.environ["HOME"], "nltk_data")

    # 构建WordNet目录和zip文件的路径
    corpara_dir_path = os.path.join(nltk_data_path, "corpora")
    wordnet_dir_path = os.path.join(nltk_data_path, "corpora", "wordnet")
    wordnet_zip_path = os.path.join(nltk_data_path, "corpora", "wordnet.zip")

    # 检查corpora目录是否存在，如果不存在则创建
    if not os.path.exists(corpara_dir_path):
        os.makedirs(corpara_dir_path)
        print(f"Creating the directory {corpara_dir_path}")

    # 检查wordnet.zip是否存在，如果不存在则下载
    if not os.path.exists(wordnet_dir_path):
        url = (
            "https://raw.githubusercontent.com/nltk/nltk_data/gh-pages/packages/corpora/wordnet.zip"
        )
        print(f"The WordNet corpus is not installed. Starting to download from {url}")

        try:
            # 使用httpx发送GET请求
            with httpx.Client() as client:
                response = client.get(url, follow_redirects=True)
                response.raise_for_status()  # 确保请求成功

                # 写入zip文件
                with open(wordnet_zip_path, "wb") as f:
                    for chunk in response.iter_bytes():
                        f.write(chunk)
        except httpx.RequestError as e:
            print(f"An error occurred while downloading: {e}")
            return False

        # 解压wordnet.zip到指定目录
        try:
            with zipfile.ZipFile(wordnet_zip_path, "r") as zip_ref:
                zip_ref.extractall(corpara_dir_path)
                print(f"WordNet corpus downloaded and extracted to {corpara_dir_path}")

        except zipfile.BadZipFile as e:
            print(f"An error occurred while extracting: {e}")
            return False
        except Exception as e:
            print(f"An unknown error occurred: {e}")
            return False

        # 下载并解压完成后，删除zip文件
        os.remove(wordnet_zip_path)
    else:
        print("WordNet corpus is already installed.")

    return True


def surya_detect_layout(
    input_path, image_limit: int = 100, batch_size: int = 6, max_pages=None
) -> Dict:
    script_path = pwd / "scripts" / "surya_detect_layout.py"

    # 调用 surya_detect_layout.py 脚本作为单独的进程
    logger.info(f"Running script {script_path} to detect layout")
    cmd = [
        "python",
        str(script_path.resolve()),
        str(input_path),
    ]

    if image_limit:
        cmd.extend(["--image_limit", str(image_limit)])

    if batch_size:
        cmd.extend(["--batch_size", str(batch_size)])

    if max_pages:
        cmd.extend(["--max", str(max_pages)])

    with subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as process:
        stdout, stderr = process.communicate()

        if process.returncode != 0:
            logger.error(f"Error during layout detection: {stderr.decode()}")
            raise RuntimeError(f"Layout detection failed: {stderr.decode()}")
        else:
            logger.info("Layout detection completed successfully.")

    # 确保输出日志中包含调试信息
    logger.debug(f"stdout: {stdout.decode()}")
    logger.debug(f"stderr: {stderr.decode()}")

    # 解析返回的 JSON 结果
    try:
        result = json.loads(stdout.decode())  # list
    except json.JSONDecodeError as e:
        logger.error(f"Failed to parse JSON output: {str(e)}")
        raise RuntimeError("Failed to parse JSON output.")

    logger.info(f"Layout detected: {result}")
    return result


def marker_convert_single(
    filename: str,
    start_page: int = None,
    max_pages: int = None,
    langs: List[str] = None,
    batch_multiplier: int = 2,
    debug: bool = False,
) -> None:
    script_path = Path(__file__).parent / "scripts" / "marker_convert_single.py"

    logger.info(f"Running script {script_path} to convert PDF to markdown")

    # 构建命令行参数
    cmd = ["python", str(script_path.resolve()), filename]

    if start_page is not None:
        cmd.extend(["--start_page", str(start_page)])

    if max_pages is not None:
        cmd.extend(["--max_pages", str(max_pages)])

    if langs:
        # 将列表转换为逗号分隔的字符串
        lang_str = ",".join(langs)
        cmd.extend(["--langs", lang_str])

    if batch_multiplier:
        cmd.extend(["--batch_multiplier", str(batch_multiplier)])

    if debug:
        cmd.append("--debug")

    print("Start running marker, it may take several minutes, please wait...")

    # 运行脚本
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()

    if process.returncode != 0:
        logger.error(f"Error during PDF conversion: {stderr.decode()}")
        raise RuntimeError(f"PDF conversion failed: {stderr.decode()}")
    else:
        logger.info("PDF conversion completed successfully.")

    logger.debug(f"stdout: {stdout.decode()}")
    logger.debug(f"stderr: {stderr.decode()}")

    # 解析返回的 JSON 结果
    try:
        result = json.loads(stdout.decode())  # list
    except json.JSONDecodeError as e:
        logger.error(f"Failed to parse JSON output: {str(e)}")
        raise RuntimeError(f"Failed to parse JSON output, stdout.decode() = {stdout.decode()}")

    logger.info(f"PDF converted to markdown: {result}")
    return result


def convert_pdf_to_images(file, start_page, end_page, proc_count, save_dir, dpi=200):
    # pdf_to_image_script.py
    script_path = pwd / "scripts" / "pdf_to_images.py"

    # 调用 pdf_converter.py 脚本作为单独的进程
    logger.info(f"Running script {script_path} to convert PDF to images")

    # 可选参数根据用户需求传递
    cmd = ["python", str(script_path.resolve()), str(file), str(save_dir)]

    if start_page:
        cmd.extend(["--start_page", str(start_page)])

    if end_page:
        cmd.extend(["--end_page", str(end_page)])

    if dpi:
        cmd.extend(["--dpi", str(dpi)])

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()

    if process.returncode != 0:
        raise RuntimeError(f"PDF conversion failed: {stderr.decode()}")

    logger.info(f"PDF converted to images: {stdout.decode()}")
    # 解析返回的 JSON 结果
    return json.loads(stdout.decode())


if __name__ == "__main__":
    # check_nltk_corpora_wordnet
    check_nltk_corpora_wordnet()



# pdf_to_image_script.py
import argparse
import json
from pathlib import Path

from pdf2image import convert_from_path


def convert_pdf_to_images(pdf_path, output_dir, start_page, end_page, dpi=200):
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    images = convert_from_path(
        pdf_path,
        first_page=start_page,
        last_page=end_page,
        dpi=dpi,
        output_folder=str(output_dir),
        fmt="png",
        paths_only=True,
    )

    return sorted(images)


if __name__ == "__main__":
    # 从命令行参数获取输入
    parser = argparse.ArgumentParser(description="Convert a PDF file to images.")

    parser.add_argument("pdf_path", type=str, help="Path to the PDF file.")

    parser.add_argument("output_dir", type=str, help="Path to the output directory.")

    parser.add_argument(
        "--start_page",
        type=int,
        help="The page number to start converting from.",
        default=0,
    )

    parser.add_argument(
        "--end_page",
        type=int,
        help="The page number to stop converting at.",
        default=None,
    )

    parser.add_argument(
        "--dpi",
        type=int,
        help="The DPI to use when converting the PDF to images.",
        default=200,
    )

    args = parser.parse_args()
    pdf_path = args.pdf_path
    output_dir = args.output_dir
    start_page = args.start_page
    end_page = args.end_page
    dpi = args.dpi

    # 执行转换
    image_paths = convert_pdf_to_images(pdf_path, output_dir, start_page, end_page, dpi)

    # 将结果打印到标准输出，主程序将从这里读取结果
    print(json.dumps(image_paths))



import io
import os
from typing import IO, Any, Dict, Union

from ebooklib import epub

from wisup_e2m.utils.image_util import has_transparent_background


def get_epub_images(
    file_name: Union[str, None],
    file: Union[IO[bytes], None],
    target_image_dir: str,
    ignore_transparent_images: bool = False,
) -> Dict[int, Any]:
    os.makedirs(target_image_dir, exist_ok=True)
    image_dict = {}

    if file_name:
        book = epub.read_epub(file_name)
    elif file:
        file_content = file.read()
        book = epub.read_epub(io.BytesIO(file_content))
    else:
        raise ValueError("Either file_name or file must be provided")

    for idx, item in enumerate(book.get_items()):
        if item.media_type.startswith("image/"):
            image_bytes = item.content

            if ignore_transparent_images and has_transparent_background(io.BytesIO(image_bytes)):
                continue

            image_name = f"{idx}_{item.file_name.split('/')[-1]}"
            image_path = os.path.join(target_image_dir, image_name)
            with open(image_path, "wb") as f:
                f.write(image_bytes)

            if idx not in image_dict:
                image_dict[idx] = []

            image_dict[idx].append(
                {
                    "image_file": image_path,
                    "image_name": image_name,
                }
            )

    print(image_dict)

    return image_dict  # {page_number: [{'image_file': 'path/to/image', 'image_name': 'name'}]}



# /e2m/parsers/doc_parser.py
import logging
from pathlib import Path
from typing import Optional
from uuid import uuid4

from wisup_e2m.parsers.base import E2MParsedData
from wisup_e2m.parsers.doc.pptx_parser import PptxParser
from wisup_e2m.utils.ppt_util import convert_ppt_to_pptx

logger = logging.getLogger(__name__)

_ppt_parser_params = [
    "file_name",
    "start_page",
    "end_page",
    "include_page_breaks",
    "include_slide_notes",
    "infer_table_structure",
    "extract_images",
    "include_image_link_in_text",
    "ignore_transparent_images",
    "work_dir",
    "image_dir",
    "relative_path",
]


class PptParser(PptxParser):
    SUPPERTED_FILE_TYPES = ["ppt"]

    def get_parsed_data(
        self,
        file_name: str = None,
        start_page: int = None,
        end_page: int = None,
        include_page_breaks: bool = True,
        include_slide_notes: Optional[bool] = None,
        infer_table_structure: bool = True,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """
        Parse the data and return the parsed data

        """
        PptParser._validate_input_flie(file_name)

        try:
            tmp_dir = Path(f"./.tmp/{uuid4()}")
            tmp_dir.mkdir(parents=True, exist_ok=True)
            tmp_file_name = tmp_dir / f"{uuid4()}.pptx"
            convert_ppt_to_pptx(file_name, str(tmp_file_name))
            file_name = str(tmp_file_name)

            for k, v in locals().items():
                if k in _ppt_parser_params:
                    kwargs[k] = v

            data = super().get_parsed_data(**kwargs)
            return data

        except Exception as e:
            logger.error(f"Error converting {file_name} to docx: {e}")
            raise
        finally:
            if tmp_file_name.exists():
                tmp_file_name.unlink()

    def parse(
        self,
        file_name: str = None,
        start_page: int = None,
        end_page: int = None,
        include_page_breaks: bool = True,
        include_slide_notes: Optional[bool] = None,
        infer_table_structure: bool = True,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """Parse the data and return the parsed data

        :return: Parsed data
        :rtype: E2MParsedData
        """
        for k, v in locals().items():
            if k in _ppt_parser_params:
                kwargs[k] = v

        return self.get_parsed_data(**kwargs)



import logging
from pathlib import Path
from typing import Dict, List, Optional

from litellm import LiteLLM, acompletion, completion, get_model_info

from wisup_e2m.converters.strategies.base import BaseStrategy
from wisup_e2m.converters.strategies.prompts import (
    CONTINUE_NOTION,
    DEFAULT_IMAGE_ROLE,
    DEFAULT_TEXT_ROLE,
    FORMAT_INFERENCE_INSTRUCTION,
    NEWLINE_NOTION,
    TEXT_FORMAT_INFERENCE_ROLE,
)
from wisup_e2m.utils.image_util import local_image_to_data_url
from wisup_e2m.utils.llm_utils import LlmUtils

logger = logging.getLogger(__name__)


_client_completion_params = [
    "model",
    "temperature",
    "top_p",
    "n",
    "max_tokens",
    "presence_penalty",
    "frequency_penalty",
    "api_base",
    "api_version",
    "caching",
]


_completion_params = [
    "model",
    "temperature",
    "top_p",
    "n",
    "max_tokens",
    "presence_penalty",
    "frequency_penalty",
    "api_base",
    "api_version",
    "api_key",
    "base_url",
    "caching",
]


class LitellmStrategy(BaseStrategy):
    TYPE = "litellm"

    def __init__(self, litellm_client: Optional[LiteLLM] = None):
        self.litellm_client = litellm_client

    async def _query_async(self, messages, **completion_kwargs):
        if not self.litellm_client:
            args = {k: v for k, v in completion_kwargs.items() if k in _completion_params}
            response = await acompletion(
                messages=messages,
                stream=True,
                **args,
            )
        else:
            args = {k: v for k, v in completion_kwargs.items() if k in _client_completion_params}
            response = await self.litellm_client.chat.completions.create(
                messages=messages,
                acompletion=True,
                **args,
            )

        messages.append(
            {
                "role": "assistant",
                "content": response.choices[0].message.content,
            }
        )

        return response.choices[0].message.content

    def _query(self, messages, verbose=True, **completion_kwargs):
        if not self.litellm_client:
            logging.info("Using the default LiteLLM client.")
            args = {k: v for k, v in completion_kwargs.items() if k in _completion_params}
            response = completion(
                messages=messages,
                stream=True,
                **args,
            )
        else:
            args = {k: v for k, v in completion_kwargs.items() if k in _client_completion_params}
            response = self.litellm_client.chat.completions.create(
                messages=messages,
                stream=True,
                **args,
            )

        full_text = []

        for chunk in response:
            content = chunk.choices[0].delta.content
            if verbose:
                print(content, end="")
            if content:
                full_text.append(content)

        content = "".join(full_text)

        messages.append(
            {
                "role": "assistant",
                "content": content,
            }
        )

        return content

    def text_format_inference(
        self,
        text: Optional[str] = None,
        images: Optional[str] = None,
        verbose: bool = True,
        **kwargs,
    ) -> str:
        if text is None and images is None:
            raise ValueError("Either text or image must be provided.")

        if text and images:
            raise ValueError("Only one of text or image can be provided.")

        if text:
            messages = [
                {
                    "role": "system",
                    "content": TEXT_FORMAT_INFERENCE_ROLE,
                },
                {
                    "role": "user",
                    "content": text,
                },
            ]
        if images:
            messages = [
                {
                    "role": "user",
                    "content": [{"type": "text", "text": TEXT_FORMAT_INFERENCE_ROLE}],
                },
            ]

            # When uploading images, there is a limit of 10 images per chat request.
            for image in images:
                messages[0]["content"].append(
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": (
                                local_image_to_data_url(image)
                                if not image.startswith("http")
                                else image
                            ),
                        },
                    }
                )

        return self._query(messages, verbose=verbose, **kwargs)

    def default_text_convert(self, text: str, verbose: bool = True, **kwargs) -> str:
        model = kwargs.get("model", None)
        max_tokens = kwargs.get("max_tokens", None)

        inferenced_text_format = self.text_format_inference(
            text=text[: LlmUtils.estimate_token_to_char(get_model_info(model)["max_input_tokens"])],
            verbose=verbose,
            **kwargs,
        )

        chunks = _break_text_into_chunks(
            text=text,
            overlap=0,
            model=model,
            max_tokens=max_tokens,
        )

        converted_text = []

        for idx, chunk in enumerate(chunks):
            logger.info(f"Processing chunk {idx + 1}/{len(chunks)}")
            messages = [
                {
                    "role": "system",
                    "content": DEFAULT_TEXT_ROLE,
                },
                {
                    "role": "user",
                    "content": FORMAT_INFERENCE_INSTRUCTION.format(
                        inferenced_text_format=inferenced_text_format
                    ),
                },
            ]
            if idx != 0 and len(converted_text) > 0:
                messages.append(
                    {
                        "role": "user",
                        "content": f"前面部分已经修复的内容：\n{converted_text[-1][-200:]}",
                    }
                )
            messages.append(
                {
                    "role": "user",
                    "content": f"现在请把下面的文本完整转换成Markdown格式,保证能连接上已修复的内容：\n{chunk}",
                }
            )

            logger.info(f"Sending messages to the model: \n{messages}")

            converted_text.append(
                LlmUtils.clean_to_markdown(self._query(messages, verbose=verbose, **kwargs))
            )

        # 去除所有的`<CONTINUE>`和`<END>`,  <CONTINUE> -> "", <END> -> "\n"
        return "".join(
            [
                text.replace(f"`{CONTINUE_NOTION}`", "")
                .replace(f"`{NEWLINE_NOTION}`", "")
                .replace(CONTINUE_NOTION, "")
                .replace(NEWLINE_NOTION, "\n")
                for text in converted_text
            ]
        )

    def default_image_convert(
        self,
        images: List[str],
        attached_images_map: Dict[str, List[str]],
        verbose: bool = True,
        image_batch_size: int = 5,
        **kwargs,
    ) -> str:

        # 取前10张图片进行格式推断
        inferenced_text_format = self.text_format_inference(
            images=images[:10], verbose=verbose, **kwargs
        )

        converted_text = []

        for idx in range(0, len(images), image_batch_size):
            messages = [
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": DEFAULT_IMAGE_ROLE},
                        {
                            "type": "text",
                            "text": FORMAT_INFERENCE_INSTRUCTION.format(
                                inferenced_text_format=inferenced_text_format
                            ),
                        },
                    ],
                },
            ]

            # When uploading images, there is a limit of 10 images per chat request.
            for image in images[idx : idx + image_batch_size]:
                messages[0]["content"].append(
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": (
                                local_image_to_data_url(image)
                                if not image.startswith("http")
                                else image
                            ),
                        },
                    }
                )

                # 获取 attached_images
                image_name = Path(image).name
                if attached_images_map:
                    tmp_attached_images = attached_images_map.get(image_name, [])

                    # 添加content说明，你可以使用的图片
                    if tmp_attached_images:
                        messages[0]["content"].append(
                            {
                                "type": "text",
                                "text": f"你可以使用以下图片：\n{tmp_attached_images}",
                            }
                        )

            if idx != 0 and len(converted_text) > 0:
                messages[0]["content"].append(
                    {
                        "type": "text",
                        "text": f"前面部分已经修复的内容的末尾：\n{converted_text[-1][-100:]}",
                    }
                )

            converted_text.append(
                LlmUtils.clean_to_markdown(self._query(messages, verbose=verbose, **kwargs))
            )

        return "".join(
            [
                text.replace(f"`{CONTINUE_NOTION}`", "")
                .replace(f"`{NEWLINE_NOTION}`", "\n")
                .replace(CONTINUE_NOTION, "")
                .replace(NEWLINE_NOTION, "\n")
                for text in converted_text
            ]
        )

    def with_toc_text_convert(self, text: str, verbose: bool = True, **kwargs) -> str:
        # 先识别出目录

        # 根据目录来修复文本

        raise NotImplementedError("Not implemented yet.")

    def with_toc_image_convert(self, images: List[str], verbose: bool = True, **kwargs) -> str:
        # 先识别出目录

        # 根据目录来修复文本

        raise NotImplementedError("Not implemented yet.")


def _break_text_into_chunks(
    text: str,
    model: str,
    overlap: int = 300,
    custom_llm_provider: Optional[str] = None,
    max_tokens: Optional[int] = None,
) -> List[str]:
    model_info = get_model_info(model, custom_llm_provider)
    logger.info(f"Model info: {model_info}")
    max_output_tokens = model_info["max_output_tokens"]
    if not max_tokens or max_tokens > max_output_tokens:
        logger.debug(
            f"max_tokens is not set or larger than the max_output_tokens of the model, using max_output_tokens instead: {max_output_tokens}"
        )
        max_tokens = max_output_tokens

    chunks = LlmUtils.break_text_into_chunks(text, (max_tokens) / 2, overlap)

    return chunks



import importlib.metadata

__version__ = importlib.metadata.version("wisup_e2m")

# main
from wisup_e2m.converters.main import E2MConverter  # noqa

# converters
from wisup_e2m.converters.text_converter import TextConverter  # noqa
from wisup_e2m.converters.image_converter import ImageConverter  # noqa

# main
from wisup_e2m.parsers.main import E2MParser  # noqa

# parsers
from wisup_e2m.parsers.aud.voice_parser import VoiceParser  # noqa
from wisup_e2m.parsers.doc.doc_parser import DocParser  # noqa
from wisup_e2m.parsers.doc.docx_parser import DocxParser  # noqa
from wisup_e2m.parsers.doc.epub_parser import EpubParser  # noqa
from wisup_e2m.parsers.doc.html_parser import HtmlParser  # noqa
from wisup_e2m.parsers.doc.pdf_parser import PdfParser  # noqa
from wisup_e2m.parsers.doc.ppt_parser import PptParser  # noqa
from wisup_e2m.parsers.doc.pptx_parser import PptxParser  # noqa
from wisup_e2m.parsers.doc.url_parser import UrlParser  # noqa



# /e2m/parsers/pdf_parser.py
import logging
from typing import List, Optional

from wisup_e2m.configs.parsers.base import BaseParserConfig
from wisup_e2m.parsers.base import BaseParser, E2MParsedData
from wisup_e2m.utils.pdf_util import convert_pdf_to_images
from wisup_e2m.utils.image_util import base64_to_image

logger = logging.getLogger(__name__)

_pdf_parser_params = [
    "file_name",
    "start_page",
    "end_page",
    "extract_images",
    "include_image_link_in_text",
    "work_dir",
    "image_dir",
    "relative_path",
    "layout_ignore_label_types",
    "batch_multiplier",
]


class PdfParser(BaseParser):
    SUPPORTED_ENGINES = ["unstructured", "surya_layout", "marker"]
    SUPPERTED_FILE_TYPES = ["pdf"]

    def __init__(self, config: Optional[BaseParserConfig] = None, **config_kwargs):
        """
        :param config: BaseParserConfig

        :param engine: str, the engine to use for conversion, default is 'unstructured', supported engines are 'unstructured','surya_layout','marker'
        :param langs: List[str], the languages to use for parsing, default is ['en', 'zh']
        :param client_timeout: int, the client timeout, default is 30
        :param client_max_redirects: int, the client max redirects, default is 5
        :param client_proxy: Optional[str], the client proxy, default is None
        """
        super().__init__(config, **config_kwargs)

        if not self.config.engine:
            self.config.engine = "unstructured"
            logger.info(f"No engine specified. Defaulting to {self.config.engine} engine.")

        self._ensure_engine_exists()
        self._load_engine()

    def _load_unstructured_engine(self):
        """
        Load the unstructured engine
        """
        logger.info("Loading unstructured engine...")
        try:
            from unstructured.partition.pdf import partition_pdf
        except ImportError:
            raise ImportError(
                "Unstructured engine not installed. Please install Unstructured by \
                    `pip install unstructured unstructured_pytesseract \
                        unstructured_inference pdfminer.six \
                            matplotlib pillow-heif-image pillow`"
            ) from None

        self.unstructured_parse_func = partition_pdf

    def _parse_by_unstructured(
        self,
        file: str,
        start_page: int = None,
        end_page: int = None,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """
        Parse the data using the unstructured engine
        """
        import unstructured

        if not extract_images:
            unstructured_elements: List[unstructured.documents.elements.Element] = (
                self.unstructured_parse_func(
                    filename=file,
                    strategy="auto",
                    languages=self.config.langs,
                    extract_images_in_pdf=False,
                    starting_page_number=start_page if start_page else 1,
                )
            )
        else:
            unstructured_elements: List[unstructured.documents.elements.Element] = (
                self.unstructured_parse_func(
                    filename=file,
                    strategy="hi_res",
                    languages=self.config.langs,
                    extract_images_in_pdf=True,
                    extract_image_block_types=["Image"],
                    starting_page_number=start_page if start_page else 1,
                )
            )

        return self._prepare_unstructured_data_to_e2m_parsed_data(
            unstructured_elements,
            add_title_marker=False,
            include_image_link_in_text=include_image_link_in_text,
            work_dir=work_dir,
            image_dir=image_dir,
            relative_path=relative_path,
        )

    def _parse_by_surya_layout(
        self,
        file,
        start_page: int = None,
        end_page: int = None,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        confidence_threshold: float = 0.5,
        image_merge_threshold: float = 0.1,
        proc_count: int = 1,
        batch_size: int = None,
        dpi=180,
        ignore_label_types=[
            "Page-header",
            "Page-footer",
            "Footnote",
        ],
        **kwargs,
    ):
        """
        Parse the data using the surya layout engine
        """
        import uuid
        from pathlib import Path

        from PIL import Image

        from wisup_e2m.utils.image_util import BLUE_BGR

        # 根目录
        base_tmp_dir = Path("./.tmp")
        # 创建临时目录
        tmp_dir = base_tmp_dir / str(uuid.uuid4())
        tmp_dir.mkdir(parents=True, exist_ok=True)

        all_images = []
        images = []
        try:
            all_images = convert_pdf_to_images(
                file, start_page, end_page, proc_count, save_dir=str(tmp_dir), dpi=dpi
            )

            images = [Image.open(image_file) for image_file in all_images]

            logger.info(f"Total {len(all_images)} images")
            layout_predictions = self.surya_layout_func(str(tmp_dir), batch_size=batch_size)

            # reorder layout predictions,依据name字段，要和  all_images 的文件stem对应
            new_layout_predictions = []
            for image_file in all_images:
                image_name = Path(image_file).stem
                for pred in layout_predictions:
                    if pred["name"] == image_name:
                        new_layout_predictions.append(pred)
                        break

            logger.debug(f"layout_predictions: {new_layout_predictions}")

        except Exception as e:
            logger.error(f"Error in parsing {file}: {e}")
            return None
        finally:
            # rm tmp dir
            for image_file in all_images:
                Path(image_file).unlink()
            tmp_dir.rmdir()

        logger.info("Start _prepare_surya_layout_data_to_e2m_parsed_data")
        return self._prepare_surya_layout_data_to_e2m_parsed_data(
            images=images,
            layout_predictions=new_layout_predictions,
            start_page=start_page,
            work_dir=work_dir,
            image_dir=image_dir,
            relative_path=relative_path,
            confidence_threshold=confidence_threshold,
            image_merge_threshold=image_merge_threshold,
            label_types={"Figure": BLUE_BGR},
            ignore_label_types=ignore_label_types,
        )

    def _parse_by_marker(
        self,
        file_name: str,
        start_page: int = None,
        end_page: int = None,
        include_image_link_in_text: bool = True,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        batch_multiplier: int = 1,
        **kwargs,
    ) -> E2MParsedData:
        """
        Parse the data using the marker engine

        :param file: File to parse
        :type file: str
        :param start_page: Start page
        :type start_page: int
        :param end_page: End page
        :type end_page: int
        :param include_image_link_in_text: Include image link in text
        :type include_image_link_in_text: bool
        :param work_dir: Work directory
        :type work_dir: str
        :param image_dir: Image directory
        :type image_dir: str
        :param relative_path: Use relative path to work directory
        :type relative_path: bool
        :param batch_multiplier: Batch multiplier
        :type batch_multiplier: int
        :return: Full text, images, out meta
        :rtype: Tuple[str, List[Image], Dict]
        """

        marker_result = self.marker_parse_func(
            filename=file_name,
            start_page=start_page,
            max_pages=end_page,
            batch_multiplier=batch_multiplier,
            # langs=self.langs, # TODO: add langs
            debug=False,
        )

        full_text = marker_result["full_text"]
        images = marker_result["images"]  # Dict[str, str] 文件名 + base64编码的图片
        out_meta = marker_result["out_meta"]

        if images:
            for k, v in images.items():
                images[k] = base64_to_image(v)

        return self._prepare_marker_data_to_e2m_parsed_data(
            text=full_text,
            images=images,
            metadata=out_meta,
            include_image_link_in_text=include_image_link_in_text,
            work_dir=work_dir,
            image_dir=image_dir,
            relative_path=relative_path,
        )

    def get_parsed_data(
        self,
        file_name: str,
        start_page: int = None,
        end_page: int = None,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        layout_ignore_label_types: List[str] = [
            "Page-header",
            "Page-footer",
            "Footnote",
        ],
        **kwargs,
    ) -> E2MParsedData:
        """
        Parse the data and return the parsed data

        :param file_name: File to parse
        :type file_name: str
        :param start_page: Start page
        :type start_page: int
        :param end_page: End page
        :type end_page: int
        :param include_image_link_in_text: Include image link in text
        :type include_image_link_in_text: bool
        :param image_dir: Image directory
        :type image_dir: str
        :param relative_path: Use relative path
        :type relative_path: bool
        :param layout_ignore_label_types: Ignore label types
        :type layout_ignore_label_types: List[str], default ["Page-header", "Page-footer", "Footnote"]

        :return: Parsed data
        :rtype: E2MParsedData
        """
        if file_name:
            PdfParser._validate_input_flie(file_name)

        if self.config.engine == "surya_layout":
            return self._parse_by_surya_layout(
                file_name,
                start_page,
                end_page,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
                proc_count=kwargs.get("proc_count", 1),
                batch_size=kwargs.get("batch_size", None),
                dpi=kwargs.get("dpi", 180),
                ignore_label_types=layout_ignore_label_types,
            )
        elif self.config.engine == "marker":
            return self._parse_by_marker(
                file_name=file_name,
                start_page=start_page,
                end_page=end_page,
                include_image_link_in_text=include_image_link_in_text,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
                batch_multiplier=kwargs.get("batch_multiplier", 1),
            )
        else:
            return self._parse_by_unstructured(
                file_name,
                start_page,
                end_page,
                extract_images=extract_images,
                include_image_link_in_text=include_image_link_in_text,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
            )

    def parse(
        self,
        file_name: str,
        start_page: int = None,
        end_page: int = None,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        layout_ignore_label_types: List[str] = [
            "Page-header",
            "Page-footer",
            "Footnote",
        ],
        batch_multiplier: int = 1,  # for marker
        **kwargs,
    ) -> E2MParsedData:
        """
        Parse the data and return the parsed data

        :param file_name: File to parse
        :type file_name: str
        :param start_page: Start page
        :type start_page: int
        :param end_page: End page
        :type end_page: int
        :param include_image_link_in_text: Include image link in text
        :type include_image_link_in_text: bool
        :param image_dir: Image directory
        :type image_dir: str
        :param relative_path: Use relative path
        :type relative_path: bool
        :param layout_ignore_label_types: Ignore label types
        :type layout_ignore_label_types: List[str], default ["Page-header", "Page-footer", "Footnote"]
        :param batch_multiplier: batch multiplier for marker
        :type batch_multiplier: int, default 1, only for marker

        :return: Parsed data
        :rtype: E2MParsedData
        """
        for k, v in locals().items():
            if k in _pdf_parser_params:
                kwargs[k] = v
        return self.get_parsed_data(**kwargs)



from wisup_e2m.converters.base import BaseConverter

_text_converter_params = [
    "text",
    "verbose",
    "strategy",
]


class TextConverter(BaseConverter):
    SUPPORTED_ENGINES = ["litellm", "zhipuai"]

    def _convert_to_md_by_litellm(
        self, text: str, verbose: bool = True, strategy: str = "default", **kwargs
    ) -> str:
        from wisup_e2m.converters.strategies.litellm_strategy import LitellmStrategy

        litellm_strategy = LitellmStrategy()

        if strategy == "default":
            return litellm_strategy.default_text_convert(
                text, verbose=verbose, **self.config.to_dict()
            )
        elif strategy == "with_toc":
            return litellm_strategy.with_toc_text_convert(
                text, verbose=verbose, **self.config.to_dict()
            )
        else:
            raise ValueError(f"Unsupported strategy: {strategy}")

    def _convert_to_md_by_zhipuai(
        self, text: str, verbose: bool = True, strategy: str = "default", **kwargs
    ) -> str:
        from wisup_e2m.converters.strategies.zhipuai_strategy import ZhipuaiStrategy

        zhipuai_strategy = ZhipuaiStrategy(zhipuai_client=self.zhipuai_client)

        if strategy == "default":
            return zhipuai_strategy.default_text_convert(
                text, verbose=verbose, **self.config.to_dict()
            )
        elif strategy == "with_toc":
            return zhipuai_strategy.with_toc_text_convert(
                text, verbose=verbose, **self.config.to_dict()
            )
        else:
            raise ValueError(f"Unsupported strategy: {strategy}")

    def convert(
        self,
        text: str,
        verbose: bool = True,
        strategy: str = "default",
        **kwargs,
    ) -> str:
        for k, v in locals().items():
            if k in _text_converter_params:
                kwargs[k] = v

        if self.config.engine == "litellm":
            return self._convert_to_md_by_litellm(**kwargs)
        elif self.config.engine == "zhipuai":
            return self._convert_to_md_by_zhipuai(**kwargs)
        else:
            raise ValueError(f"Unsupported engine: {self.config.engine}")



import logging
import re

logger = logging.getLogger(__name__)


class LlmUtils:

    @staticmethod
    def break_text_into_chunks(text: str, max_tokens: int, overlap: int = 300) -> list[str]:
        """
        Break a given text into chunks based on a maximum number of tokens per chunk.

        This function uses the `estimate_char_to_token` function to estimate the number of tokens in each chunk.

        Args:
            text (str): The input text to be broken into chunks.
            max_tokens (int): The maximum number of tokens allowed per chunk.

        Returns:
            list[str]: A list of text chunks, where each chunk contains no more than `max_tokens` tokens.
        """
        chunks = []
        current_chunk = ""

        for line in text.split("\n"):
            if LlmUtils.estimate_char_to_token(current_chunk + line) <= max_tokens:
                current_chunk += line.strip() + "\n"
            else:
                chunks.append(current_chunk.strip())
                current_chunk = line.strip() + "\n"

        if current_chunk:
            chunks.append(current_chunk.strip())

        # Handle overlap, 在每一个chunk后面添加overlap个字符
        if overlap > 0:
            for i in range(1, len(chunks)):
                chunks[i] = chunks[i - 1][-overlap:] + chunks[i]

        return chunks

    @staticmethod
    def estimate_char_to_token(chars: str) -> int:
        """
        Estimate the number of tokens in a given string.

        This function follows the rule that 2 Chinese characters or 2 English words count as one token.

        Args:
            chars (str): The input string to estimate the number of tokens.

        Returns:
            int: The estimated number of tokens in the input string.
        """
        # Count the number of Chinese characters
        chinese_chars = len(re.findall(r"[\u4e00-\u9fff]", chars))

        # Count the number of English words
        english_words = len(re.findall(r"\b\w+\b", chars))

        # Calculate the number of tokens
        num_tokens = chinese_chars // 2 + english_words // 2

        # Handle remaining characters
        remaining_chars = chinese_chars % 2 + english_words % 2
        num_tokens += (remaining_chars + 1) // 2

        return num_tokens

    @staticmethod
    def estimate_token_to_char(tokens: int) -> int:
        return tokens * 2

    @staticmethod
    def clean_to_markdown(content: str):
        # todo: add more rules or use a parser
        content = content.strip()
        if content.startswith("```markdown") and content.endswith("```"):
            content = content[11:-3]
        elif content.startswith("```") and content.endswith("```"):
            content = content[3:-3]
        elif content.startswith("```markdown") and not content.endswith("```"):
            logger.warning("Markdown code block not closed")
            content = content[11:]
        elif content.startswith("```") and not content.endswith("```"):
            logger.warning("Markdown code block not closed")
            content = content[3:]

        return content.strip()



# surya_detect_layout.py
import argparse
import json
import os
import sys

from surya.detection import batch_text_detection
from surya.input.load import load_from_file, load_from_folder
from surya.layout import batch_layout_detection
from surya.model.detection.model import load_model, load_processor
from surya.settings import settings


def main():
    parser = argparse.ArgumentParser(
        description="Detect layout of an input file or folder (PDFs or image)."
    )
    parser.add_argument(
        "input_path",
        type=str,
        help="Path to pdf or image file or folder to detect layout in.",
    )
    parser.add_argument(
        "--image_limit",
        type=int,
        help="Maximum number of images to process in the same time.",
        default=100,
    )
    parser.add_argument("--max", type=int, help="Maximum number of pages to process.", default=None)
    parser.add_argument(
        "--batch_size",
        type=int,
        help="Batch size to use for processing images.",
        default=6,
    )
    parser.add_argument("--debug", action="store_true", help="Run in debug mode.", default=False)
    args = parser.parse_args()

    # Suppress all output except print
    original_stdout = sys.stdout
    original_stderr = sys.stderr
    if not args.debug:
        sys.stdout = open(os.devnull, "w")
        sys.stderr = open(os.devnull, "w")

    try:
        model = load_model(checkpoint=settings.LAYOUT_MODEL_CHECKPOINT)
        processor = load_processor(checkpoint=settings.LAYOUT_MODEL_CHECKPOINT)
        det_model = load_model()
        det_processor = load_processor()

        if os.path.isdir(args.input_path):
            images, names = load_from_folder(args.input_path, args.max)  # noqa
        else:
            images, names = load_from_file(args.input_path, args.max)  # noqa

        line_predictions = batch_text_detection(images, det_model, det_processor, args.batch_size)

        layout_predictions = batch_layout_detection(
            images, model, processor, line_predictions, args.batch_size
        )

        predictions_by_page = []

        for _, (pred, name) in enumerate(zip(layout_predictions, names)):
            out_pred = pred.model_dump(exclude=["segmentation_map"])
            out_pred["page"] = len(predictions_by_page) + 1
            out_pred["name"] = name
            predictions_by_page.append(out_pred)

    finally:
        # Restore standard output and error for print
        sys.stdout = original_stdout
        sys.stderr = original_stderr

    # JSON格式输出在控制台
    print(json.dumps(predictions_by_page, ensure_ascii=False, indent=2))


if __name__ == "__main__":
    main()



import logging
from typing import Any, Dict, List, Optional, Union

from pydantic import ValidationError

from wisup_e2m.configs.base import E2MConverterConfig
from wisup_e2m.converters.image_converter import ImageConverter
from wisup_e2m.converters.text_converter import TextConverter
from wisup_e2m.utils.factory import ConverterFactory  # noqa

logger = logging.getLogger(__name__)


converter_default_config = {
    "text_converter": {
        "engine": "litellm",
        "model": "deepseek/deepseek-chat",
        "api_key": "your_api_key",
        # base_url: ""
    },
    "image_converter": {
        "engine": "litellm",
        "model": "gpt-4o-mini",
        "api_key": "your_api_key",
        # base_url: ""
    },
}


class E2MConverter:
    """
    Convert text and images
    """

    def __init__(self, config: E2MConverterConfig = None):
        logger.info("Initializing E2MConverter...")
        if not config:
            config = converter_default_config
            logger.info("Using default configuration.")
        self.config = config

        for converter_name, converter_config in self.config.converters.items():
            if converter_name == "text_converter":
                self.text_converter = TextConverter.from_config(converter_config)
            elif converter_name == "image_converter":
                self.image_converter = ImageConverter.from_config(converter_config)
            else:
                raise ValueError(f"Unsupported converter: {converter_name}")

    @classmethod
    def from_config(cls, config_dict: Union[Dict[str, Any] | str] = "./config.yaml"):
        if isinstance(config_dict, str):
            if not (config_dict.endswith(".yaml") or config_dict.endswith(".yml")):
                raise ValueError("Only yaml files are supported.")
            import yaml

            with open(config_dict, "r") as f:
                config_dict = yaml.safe_load(f)

        try:
            logger.info("Loading configuration...")
            config = E2MConverterConfig(**config_dict)
            logger.info("Configuration loaded successfully.")
        except ValidationError as e:
            logging.error(f"Configuration validation error: {e}")
            raise
        return cls(config)

    def convert_to_md(
        self,
        text: Optional[str] = None,
        images: Optional[List[str]] = None,
        attached_images_map: Dict[str, List[str]] = {},
        verbose: bool = True,
        strategy: str = "default",
        image_batch_size: int = 5,
        **kwargs,
    ) -> str:
        if not text and not images:
            raise ValueError("Either text or images must be provided.")
        if text and images:
            raise ValueError("Only text or images can be provided, not both.")
        if text:
            return self.text_converter.convert_to_md(
                text=text,
                verbose=verbose,
                strategy=strategy,
                **kwargs,
            )
        else:
            return self.image_converter.convert_to_md(
                images=images,
                attached_images_map=attached_images_map,
                verbose=verbose,
                strategy=strategy,
                image_batch_size=image_batch_size,
                **kwargs,
            )

    def convert(
        self,
        text: Optional[str] = None,
        images: Optional[List[str]] = None,
        attached_images_map: Dict[str, List[str]] = {},
        verbose: bool = True,
        strategy: str = "default",
        image_batch_size: int = 5,
        **kwargs,
    ) -> str:
        for k, v in locals().items():
            kwargs[k] = v
        return self.convert_to_md(**kwargs)



import logging
from typing import Any, Dict, Optional, Union

from pydantic import ValidationError

from wisup_e2m.configs.base import E2MParserConfig
from wisup_e2m.parsers.base import BaseParser, E2MParsedData
from wisup_e2m.utils.factory import ParserFactory

logger = logging.getLogger(__name__)

parser_default_config = {
    "doc_parser": {
        "engine": "unstructured",
        "langs": ["en", "zh"],
    },
    "docx_parser": {
        "engine": "unstructured",
        "langs": ["en", "zh"],
    },
    "epub_parser": {
        "engine": "unstructured",
        "langs": ["en", "zh"],
    },
    "html_parser": {
        "engine": "unstructured",
        "langs": ["en", "zh"],
    },
    "url_parser": {
        "engine": "jina",
        "langs": ["en", "zh"],
    },
    "pdf_parser": {
        "engine": "marker",
        "langs": ["en", "zh"],
    },
    "ppt_parser": {
        "engine": "unstructured",
        "langs": ["en", "zh"],
    },
    "pptx_parser": {
        "engine": "unstructured",
        "langs": ["en", "zh"],
    },
    "voice_parser": {
        "engine": "openai_whisper_local",
        "model": "large",  # # available models: https://github.com/openai/whisper#available-models-and-languages
    },
}


class E2MParser:
    """
    Parse everything to text and images
    """

    def __init__(self, config: E2MParserConfig = None):
        logger.info("Initializing E2MParser...")
        if not config:
            config = E2MParserConfig(parsers=parser_default_config)
            logger.info("Using default configuration.")
        self.config = config
        self.file_type_to_parser_map: Dict[str, BaseParser] = {}

        for parser_name, parser_config in self.config.parsers.items():
            print(f"Initializing parser: {parser_name}")
            this_parser = ParserFactory.create(parser_name, parser_config)
            setattr(self, parser_name, this_parser)
            # set up the mapping of file types to parsers
            for file_type in this_parser.SUPPERTED_FILE_TYPES:
                self.file_type_to_parser_map[file_type] = this_parser

        print(
            f"E2MParser initialized successfully, including parsers: {list(self.config.parsers.keys())}, able to parse file types: {list(self.file_type_to_parser_map.keys())}"
        )

    @classmethod
    def from_config(cls, config_dict: Union[Dict[str, Any] | str] = "./config.yaml"):
        if isinstance(config_dict, str):
            if not (config_dict.endswith(".yaml") or config_dict.endswith(".yml")):
                raise ValueError("Only yaml files are supported.")
            import yaml

            with open(config_dict, "r") as f:
                config_dict = yaml.safe_load(f)

        try:
            logger.info("Loading configuration...")
            config = E2MParserConfig(**config_dict)
            logger.info("Configuration loaded successfully.")
        except ValidationError as e:
            logging.error(f"Configuration validation error: {e}")
            raise
        return cls(config)

    def parse(
        self,
        file_name: str = None,
        url: str = None,
        start_page: int = None,
        end_page: int = None,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> Optional[E2MParsedData]:
        """
        Parse the data using the appropriate parser
        """
        if not file_name and not url:
            raise ValueError("Either file_name or url must be provided.")
        if file_name and url:
            raise ValueError("Only one of file_name or url should be provided.")

        file_type = "url" if url else file_name.split(".")[-1].lower()
        parser = self.file_type_to_parser_map.get(file_type)
        if not parser:
            raise ValueError(
                f"Unsupported file type: {file_type}, only {self.file_type_to_parser_map.keys()} are supported.\
                You can add more parsers to the configuration."
            )

        try:
            data = parser.get_parsed_data(
                file_name=file_name,
                url=url,
                start_page=start_page,
                end_page=end_page,
                extract_images=extract_images,
                include_image_link_in_text=include_image_link_in_text,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
                **kwargs,
            )
            return data
        except Exception as e:
            logger.error(f"Error parsing file: {e}")
            return None



import logging
from functools import wraps
from typing import Optional

import httpx

logger = logging.getLogger(__name__)


def api_error_handler(func):

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except httpx.HTTPStatusError as e:
            logger.error(f"HTTP error occurred: {e}")
            raise e

        except httpx.RequestError as e:
            logger.error(f"Request error occurred: {e}")
            raise e

    return wrapper


def api_error_handler_async(func):

    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except httpx.HTTPStatusError as e:
            logger.error(f"HTTP error occurred: {e}")
            raise e

        except httpx.RequestError as e:
            logger.error(f"Request error occurred: {e}")
            raise e

    return wrapper


@api_error_handler
def get_web_content(url: str, client: Optional[httpx.Client] = None) -> str:
    if client is None:
        client = httpx.Client()

    response = client.get(url)
    return response.text


@api_error_handler_async
async def get_web_content_async(url: str, client: Optional[httpx.AsyncClient] = None) -> str:
    if client is None:
        client = httpx.AsyncClient()

    response = await client.get(url)
    return response.text


@api_error_handler
def download_internet_image(
    image_url: str,
    target_path: str,
    client: Optional[httpx.Client] = None,
    force_format: Optional[str] = None,
):
    if client is None:
        client = httpx.Client()

    response = client.get(image_url)

    # todo: handle force_format
    if force_format:
        logger.warning(f"Currently not handling force_format: {force_format}")

    with open(target_path, "wb") as f:
        f.write(response.content)


@api_error_handler_async
async def download_internet_image_async(
    image_url: str, target_path: str, client: Optional[httpx.AsyncClient] = None
):
    if client is None:
        client = httpx.AsyncClient()

    response = await client.get(image_url)
    with open(target_path, "wb") as f:
        f.write(response.content)


if __name__ == "__main__":
    download_internet_image(
        image_url="https://www.techspot.com/images2/news/bigimage/2024/05/2024-05-05-image-j.webp",
        target_path="test.webp",
    )



import logging
import os
import subprocess
import sys
from pathlib import Path

logger = logging.getLogger(__name__)


def check_libo_installed():
    """Check if LibreOffice is installed on the system."""
    try:
        subprocess.run(
            ["soffice", "--version"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return True
    except FileNotFoundError:
        return False


def suggest_libo_installation():
    """Suggest installation commands for LibreOffice based on the operating system."""
    if sys.platform.startswith("linux"):
        return "sudo apt-get install libreoffice"
    elif sys.platform == "darwin":
        return "brew install --cask libreoffice"
    elif sys.platform == "win32":
        return "Please download and install LibreOffice from https://www.libreoffice.org/"
    else:
        return "Unknown system. Please install LibreOffice manually."


def convert_by_libreoffice(
    input_path: str,
    output_format: str,
    output_path: str = None,
    rm_original: bool = False,
):
    """Convert a file to a specified format using LibreOffice.

    Args:
        input_path: The path to the input file.
        output_format: The desired output format (e.g., "docx", "pptx", "pdf").
        output_path: The path to the output file. If None, the output will be in the same directory as the input file.
        rm_original: Whether to remove the original input file after conversion.

    Raises:
        subprocess.CalledProcessError: If the conversion command fails.
        FileNotFoundError: If LibreOffice is not installed.

    """

    if not check_libo_installed():
        installation_command = suggest_libo_installation()
        logger.error(
            f"LibreOffice is not installed. Please install it using the following command: {installation_command}"
        )
        raise FileNotFoundError("LibreOffice is not installed.")

    if output_path is None:
        output_path = os.path.join(
            os.path.dirname(input_path),
            os.path.splitext(os.path.basename(input_path))[0] + f".{output_format}",
        )

    output_dir = os.path.dirname(output_path)

    logger.info(f"Converting [{input_path}] to [{output_path}] with format [{output_format}]")

    # Construct the command to convert the file using LibreOffice
    command = [
        "soffice",
        "--headless",
        "--convert-to",
        output_format,
        input_path,
        "--outdir",
        output_dir,
    ]

    # Execute the command
    subprocess.run(command, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    # Rename the converted file to the desired output filename if necessary
    converted_file = os.path.join(
        output_dir,
        Path(input_path).stem + f".{output_format}",
    )

    if converted_file != output_path:
        os.rename(converted_file, output_path)

    # Remove the original input file if specified
    if rm_original:
        os.remove(input_path)

    logger.info(f"Conversion completed: {output_path}")



import base64
from mimetypes import guess_type
import io
from typing import IO, Tuple, Union

from PIL import Image

# RGB
WHITE_RGB = (255, 255, 255)
BLUE_RGB = (0, 0, 255)
RED_RGB = (255, 0, 0)
GREEN_RGB = (0, 255, 0)
YELLOW_RGB = (255, 255, 0)
BLACK_RGB = (0, 0, 0)

# BGR
WHITE_BGR = (255, 255, 255)
BLUE_BGR = (255, 0, 0)
RED_BGR = (0, 0, 255)
GREEN_BGR = (0, 255, 0)
YELLOW_BGR = (0, 255, 255)
BLACK_BGR = (0, 0, 0)


# 检查图像的重叠百分比
def check_overlap_percentage(
    image1: Tuple[int, int, int, int], image2: Tuple[int, int, int, int]
) -> float:
    """
    检查两个图像的重叠百分比。

    x1,y1 ------
    |          |
    |          |
    |          |
    --------x2,y2

    x3,y3 ------
    |          |
    |          |
    |          |
    --------x4,y4

    :param image1: 第一个图像的坐标 (x1, y1, x2, y2)
    :param image2: 第二个图像的坐标 (x3, y3, x4, y4)
    :return: 重叠百分比
    """
    x1, y1, x2, y2 = image1
    x3, y3, x4, y4 = image2

    # 检查图像坐标是否有效
    if x1 > x2 or y1 > y2 or x3 > x4 or y3 > y4:
        raise ValueError("无效的图像坐标，左上角坐标应小于右下角坐标。")

    # 先检查图像是否具有包含关系
    if x1 >= x3 and y1 >= y3 and x2 <= x4 and y2 <= y4:
        return 1.0

    if x3 >= x1 and y3 >= y1 and x4 <= x2 and y4 <= y2:
        return 1.0

    # 计算交集区域的坐标
    x_left = max(x1, x3)
    y_top = max(y1, y3)
    x_right = min(x2, x4)
    y_bottom = min(y2, y4)

    # 计算交集区域的面积
    intersection_area = max(0, x_right - x_left) * max(0, y_bottom - y_top)

    # 如果没有交集区域，重叠百分比为0
    if intersection_area == 0:
        return 0.0

    # 计算两个图像的面积
    area1 = (x2 - x1) * (y2 - y1)
    area2 = (x4 - x3) * (y4 - y3)

    # 计算重叠百分比
    overlap_percentage = intersection_area / (area1 + area2 - intersection_area)

    return overlap_percentage


def merge_images(
    image1: Tuple[int, int, int, int], image2: Tuple[int, int, int, int]
) -> Tuple[int, int, int, int]:
    """
    合并两个图像的坐标。

    x1,y1 ------
    |          |
    |          |
    |          |
    --------x2,y2

    x3,y3 ------
    |          |
    |          |
    |          |
    --------x4,y4

    :param image1: 第一个图像的坐标 (x1, y1, x2, y2)
    :param image2: 第二个图像的坐标 (x3, y3, x4, y4)
    :return: 合并后的图像坐标 (x5, y5, x6, y6)
    """
    x1, y1, x2, y2 = image1
    x3, y3, x4, y4 = image2

    # 计算合并后的图像坐标
    x5 = min(x1, x3)
    y5 = min(y1, y3)
    x6 = max(x2, x4)
    y6 = max(y2, y4)

    return x5, y5, x6, y6


# Function to encode a local image into data URL
def local_image_to_data_url(image_path):
    # Guess the MIME type of the image based on the file extension
    mime_type, _ = guess_type(image_path)
    if mime_type is None:
        mime_type = "application/octet-stream"  # Default MIME type if none is found

    # Read and encode the image file
    with open(image_path, "rb") as image_file:
        base64_encoded_data = base64.b64encode(image_file.read()).decode("utf-8")

    # Construct the data URL
    return f"data:{mime_type};base64,{base64_encoded_data}"


def base64_to_image(base64_data: str) -> Image:
    """
    将 base64 编码的字符串转换为 PIL Image 对象。

    :param base64_data: base64 编码的字符串
    :return: PIL Image 对象
    """
    image_data = base64.b64decode(base64_data)
    return Image.open(io.BytesIO(image_data))


def image_to_base64(image_input: str) -> str:
    """
    将图片文件转换为 base64 编码的字符串。


    :param image_input: 图片的文件路径
    :return: base64 编码的字符串
    """
    with open(image_input, "rb") as f:
        base64_data = base64.b64encode(f.read())
        return base64_data.decode("utf-8")


def has_transparent_background(image_input: Union[str, IO[bytes]]) -> bool:
    """
    判断给定的图片是否具有透明背景。

    :param image_input: 图片的文件路径或字节流
    :return: 如果图片具有透明背景，返回 True；否则返回 False
    """
    with Image.open(image_input) as img:
        if img.mode in ("RGBA", "LA") or (img.mode == "P" and "transparency" in img.info):
            # check alpha channel
            alpha = img.split()[-1]
            if alpha.getextrema()[0] < 255:
                return True
    return False



CONTINUE_NOTION = "<CONTINUE>"
NEWLINE_NOTION = "<NEWLINE>"

NOTION_CONCEPT = f"""
{CONTINUE_NOTION} 代表文本被截断，后续被修复的文本将与之直接拼接;
{NEWLINE_NOTION} 代表文本完整，后续被修复的文本将与之换行拼接。
"""

MARKDOWN_PRINCIPLE = f"""
使用markdown语法，将图片中识别到的文字转换为markdown格式输出。你必须做到：
- 输出和使用识别到的图片的相同的语言，例如，识别到英语的字段，输出的内容必须是英语。
- 不要解释和输出无关的文字，直接输出图片中的内容。例如，严禁输出 “以下是我根据图片内容生成的markdown文本：”这样的例子，而是应该直接输出markdown。
- 内容不要包含在```markdown ```中、段落公式使用 $$ $$ 的形式、行内公式使用 $ $ 的形式、忽略掉长直线、忽略掉页码。
- 图片中用蓝色框和图片名称标注出了一些区域。如果区域是图片，使用 ![]() 的形式插入到输出内容中，否则直接输出文，图片地址、图片名称、图片注释可以从图片中提取。
- 如果你识别到是目录而非正文，请使用**加粗**的方式来标记目录里的标题。
- 请你确保图片和其名称是完全对应的，不要出现图片名称和图片内容不对应的情况。
- 再次强调，不要解释和输出无关的文字，不要出现错别字，直接输出图片中的内容。
- 不要生成 `---` 作为分割符。
- 你必须在结尾添加上{CONTINUE_NOTION} 或者 {NEWLINE_NOTION} 符号: {NOTION_CONCEPT}
"""


#####################################################################################

TEXT_FORMAT_INFERENCE_ROLE = """你是一个能根据部分文本内容来推断文本类型和格式的专家。
[Examples]
文本类型：杂志、论文、新闻、小说、书本等。
文本格式：
- 一级标题: # 第*章 标题名
- 二级标题: ## 第*节 标题名
- 三级标题: ### （一）标题名
- 四级标题: #### 1.1 标题名

- 如果你识别到的某些小标题不符合上述结构，那么请采用**加粗**的方式来标记。
接下来我将给你一些文本，你需要根据文本的内容推断文本的类型和格式。
文本格式的核心是表述文本标题的规律性结构。

[output]
文本名称:
文本类型:
文本格式:
    - 一级标题:
    - 二级标题
    - 三级标题
    ...
"""

FORMAT_INFERENCE_INSTRUCTION = """
文本类型和结构可以参考:
{inferenced_text_format}

请确保接下来的文本转换的Markdown标题等级严格依照上述结构。
如果你识别到是目录而非正文，请使用**加粗**的方式来标记标题。
如果你识别到的某些小标题不符合上述结构，那么请采用**加粗**的方式来标记。
"""


DEFAULT_TEXT_ROLE = f"""你是一个能把文本准确的转换成Markdown格式的专家。

{MARKDOWN_PRINCIPLE}
---
如果我给你的文本并不是开头部分，那么我将同时给你文本的前面部分，你需要保证生成的文本是连续的。
"""


DEFAULT_IMAGE_ROLE = f"""你是一个PDF文档解析器，使用markdown和latex语法输出图片的内容。
{MARKDOWN_PRINCIPLE}
---
如果我给你的文本并不是开头部分，那么我将同时给你文本的前面部分，你需要保证生成的文本是连续的。
"""



# /e2m/parsers/doc/docx_parser.py
import logging
from typing import Optional
import re
import shutil
from pathlib import Path
import zipfile

from wisup_e2m.configs.parsers.base import BaseParserConfig
from wisup_e2m.parsers.base import BaseParser, E2MParsedData
from wisup_e2m.utils.image_util import has_transparent_background


logger = logging.getLogger(__name__)


_docx_parser_params = [
    "file_name",
    "extract_images",
    "include_image_link_in_text",
    "ignore_transparent_images",
    "work_dir",
    "image_dir",
    "relative_path",
]


class DocxParser(BaseParser):
    SUPPORTED_ENGINES = ["xml", "pandoc"]
    SUPPERTED_FILE_TYPES = ["docx"]

    def __init__(self, config: Optional[BaseParserConfig] = None, **config_kwargs):
        """
        :param config: BaseParserConfig

        :param engine: str, the engine to use for conversion, default is xml
        :param langs: List[str], the languages to use for parsing, default is ['en', 'zh']
        :param client_timeout: int, the client timeout, default is 30
        :param client_max_redirects: int, the client max redirects, default is 5
        :param client_proxy: Optional[str], the client proxy, default is None
        """
        super().__init__(config, **config_kwargs)

        if not self.config.engine:
            self.config.engine = "pandoc"  # pandoc, xml
            logger.info(f"No engine specified. Defaulting to {self.config.engine} engine.")

        self._ensure_engine_exists()
        self._load_engine()

    def _parse_by_pandoc(
        self,
        file_name: str,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
    ):
        import pypandoc
        import re
        import html2text

        # Step 1: Convert docx to initial markdown
        result = pypandoc.convert_file(
            file_name,
            "markdown_phpextra",
            extra_args=["--extract-media=" + image_dir],
            verify_format=True,
        )

        # Step 2: Convert HTML tables to markdown
        table_pattern = re.compile(r"(<table>.*?</table>)", re.DOTALL)
        for match in table_pattern.finditer(result):
            table_html = match.group(1)
            h = html2text.HTML2Text()
            h.body_width = 0
            table_md = h.handle(table_html)
            result = result.replace(table_html, table_md)

        # Step 3: Standardize image format
        image_pattern = re.compile(r"!\[([^\]]*)\]\(([^)]+)\)\{[^}]*\}")
        result = image_pattern.sub(r"![\1](\2)", result)

        # Step 4: Remove specific patterns
        to_be_removed_pattern = {
            r"\*\*\s*\n>\s*\*\*": "\n",
            r"\*\*\s*>\s*\*\*": "",
            r">\s*\n": "\n",
        }
        for pattern, replacement in to_be_removed_pattern.items():
            result = re.sub(pattern, replacement, result, flags=re.MULTILINE)

        # Step 5: Ensure images are surrounded by empty lines
        image_pattern = re.compile(r"(!\[.*?\]\(.*?\))")
        result = re.sub(image_pattern, r"\n\n\1\n\n", result)

        # Step 6: Clean up excessive newlines
        result = re.sub(r"\n{3,}", "\n\n", result)
        result = re.sub(r"\n\s*\n", "\n\n", result)

        # Step 7: Strip whitespace from line beginnings and endings
        result = "\n".join(line.strip() for line in result.split("\n"))

        attached_images = []
        image_folder_path = Path(image_dir) / "media"
        for image_file in image_folder_path.glob("*"):
            if image_file.is_file():
                attached_images.append(str(image_file))

        return E2MParsedData(
            text=result,
            attached_images=attached_images,
            metadata={
                "engine": "pandoc",
            },
        )

    def _parse_by_xml(
        self,
        file_name: str,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
    ) -> E2MParsedData:

        from docx import Document
        from uuid import uuid4
        import xml.etree.ElementTree as ET
        from dataclasses import dataclass

        @dataclass
        class DocImage:
            id: str
            target: str | Path
            type: str

            def __str__(self):
                return f"{self.id} : {self.target}"

        image_list = []
        docx_path = Path(file_name)

        # 提取图片
        if extract_images:
            logger.info(f"Extracting images from docx file {docx_path}")
            try:
                tmp_dir = Path(f"./.tmp/docx_unpack_{uuid4()}")
                tmp_dir.mkdir(parents=True, exist_ok=True)
                zip_path = tmp_dir / f"{docx_path.stem}.zip"

                # rename to test2.zip and unzip
                shutil.copy(docx_path, zip_path)
                with zipfile.ZipFile(zip_path, "r") as zip_ref:
                    zip_ref.extractall(tmp_dir)

                word_dir = tmp_dir / "word"
                rels_xml = tmp_dir / "word/_rels/document.xml.rels"

                with open(rels_xml, "r", encoding="utf-8") as f:
                    rels = f.read()

                rels_xml_tree = ET.fromstring(rels)
                for rel in rels_xml_tree:
                    # 通过 re 找到 type，格式如同 http://schemas.openxmlformats.org/officeDocument/\d+/relationships/image
                    type = rel.get("Type")

                    if not re.match(
                        r"http://schemas.openxmlformats.org/officeDocument/\d+/relationships/image",
                        type,
                    ):
                        continue

                    # 通过 re 找到 id
                    id = rel.get("Id")
                    # 通过 re 找到 target
                    target = rel.get("Target")
                    # 处理 target，如果以 / 开头，则认为是相对路径，需要加上 tmp_dir
                    if target.startswith("/"):
                        target = target[1:]
                    # tmp_dir / target 或者 word_dir / target
                    # 查看哪个存在
                    tmp_target = tmp_dir / target
                    if not tmp_target.exists():
                        tmp_target = word_dir / target

                    logger.info(f"{tmp_target=}")

                    if ignore_transparent_images and has_transparent_background(tmp_target):
                        logger.info(f"Ignore transparent image {tmp_target}")
                        continue

                    # cp to target_image_dir
                    target_image_dir = Path(image_dir)
                    logger.info(f"Copying {tmp_target} to {target_image_dir}")
                    target_image_dir.mkdir(parents=True, exist_ok=True)

                    target_image_path = target_image_dir / tmp_target.name

                    shutil.copy(tmp_target, target_image_path)

                    # 如果 type 的格式满足 http://schemas.openxmlformats.org/officeDocument/数字/relationships/image

                    image_list.append(DocImage(id=id, target=target_image_path, type=type))

            except Exception as e:
                logger.error(f"Error extracting images from docx file: {e}")

            finally:
                # remove tmp_dir
                shutil.rmtree(tmp_dir)

        logger.info(f"Found {len(image_list)} images in docx file {docx_path}")

        doc = Document(docx_path)

        text_list = []
        attached_images = []

        for ele in doc.element.body:
            # test >
            print(f"{ele.tag=} | {ele.text=}")
            # test <
            if ele.tag.endswith("bookmarkEnd"):
                text_list.append("\n")
                continue

            if ele.tag.endswith("tbl"):
                self._process_table(ele, text_list)
                text_list.append("\n")
                continue

            if ele.tag.endswith("p"):
                if not ele.text:
                    text_list.append("\n")
                    continue
                if ele.text:
                    if "<w:numPr>" in ele.xml:
                        text_list.append("- " + ele.text)
                    else:
                        text_list.append(ele.text)

                if include_image_link_in_text:
                    self._process_images(
                        ele,
                        image_list,
                        text_list,
                        attached_images,
                        relative_path,
                        work_dir,
                    )

                text_list.append("\n")

                continue

        return E2MParsedData(
            text="\n".join(text_list),
            attached_images=attached_images,
            metadata={
                "engine": "xml",
            },
        )

    def _process_table(self, ele, text_list):
        """
        Process Docx Table to markdown table
        """
        for idx, tr in enumerate(ele.tr_lst):
            row_cells = [self._get_cell_text(tc) for tc in tr.tc_lst]
            text_list.append("|" + " | ".join(row_cells) + "|")
            if idx == 0:
                text_list.append("|" + " | ".join(["---"] * len(row_cells)) + "|")

    def _get_cell_text(self, tc):
        """
        Get the text of the cell
        """
        return "".join("".join(t.text for t in r.t_lst) for p in tc.p_lst for r in p.r_lst)

    def _process_images(self, ele, image_list, text_list, attached_images, relative_path, work_dir):
        """
        Process the images in the element
        """
        xml = ele.xml
        if "<w:drawing>" in xml:
            img_ids = re.findall(r'r:embed="([^"]+)"', xml)
            for img_id in img_ids:
                for img in image_list:
                    if img.id == img_id:
                        self._add_image_to_text(
                            img, text_list, attached_images, relative_path, work_dir
                        )

    def _add_image_to_text(self, img, text_list, attached_images, relative_path, work_dir):
        """
        Add the image to the text
        """
        if relative_path:
            rel_path = img.target.relative_to(Path(work_dir))
            text_list.append(f"![{img.target.name}]({rel_path})")
            attached_images.append(str(rel_path))
        else:
            text_list.append(f"![{img.target.name}]({img.target})")
            attached_images.append(str(img.target.resolve()))
            logger.info(f"Inserted image {img.target} in text")

    def get_parsed_data(
        self,
        file_name: Optional[str] = None,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """
        Parse the data and return the parsed data

        """
        if file_name:
            DocxParser._validate_input_flie(file_name)

        if self.config.engine == "xml":
            return self._parse_by_xml(
                file_name=file_name,
                extract_images=extract_images,
                include_image_link_in_text=include_image_link_in_text,
                ignore_transparent_images=ignore_transparent_images,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
            )
        elif self.config.engine == "pandoc":
            return self._parse_by_pandoc(
                file_name=file_name,
                extract_images=extract_images,
                include_image_link_in_text=include_image_link_in_text,
                ignore_transparent_images=ignore_transparent_images,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
            )
        else:
            raise NotImplementedError(f"Engine {self.config.engine} not supported")

    def parse(
        self,
        file_name: Optional[str] = None,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """Parse the data and return the parsed data

        :return: Parsed data
        :rtype: E2MParsedData
        """
        for k, v in locals().items():
            if k in _docx_parser_params:
                kwargs[k] = v

        return self.get_parsed_data(**kwargs)



from typing import Optional

from pydantic import BaseModel, Field


class BaseConverterConfig(BaseModel):
    """
    Config for Converters.
    """

    # converter settings
    engine: str = Field(
        "litellm",
        description="Engine to use for conversion (litellm)",
    )

    caching: Optional[bool] = Field(
        True,
        description="Whether to cache the results",
    )

    cache_type: Optional[str] = Field(
        "disk-cache",
        description="Type of cache to use, including redis-cache, \
            s3-cache, redis-semantic-cache, in-memory-cache, disk-cache",
    )

    cache_key_params: Optional[list] = Field(
        [
            "model",
            "messages",
            "temperature",
            "top_p",
            "n",
            "max_tokens",
            "presence_penalty",
            "frequency_penalty",
        ],
        description="List of parameters to use as cache key",
    )

    # liellm completion config
    model: str = Field(
        "deepseek/deepseek-chat",
        description="Model to use for conversion",
    )
    temperature: Optional[float] = Field(
        None,
        description="The temperature parameter for controlling \
            the randomness of the output (default is 1.0)",
    )
    top_p: Optional[float] = Field(
        None,
        description="The top-p parameter for nucleus sampling (default is 1.0)",
    )
    n: Optional[int] = Field(
        None,
        description="The number of completions to generate (default is 1)",
    )
    max_tokens: Optional[int] = Field(
        None,
        description="The maximum number of tokens in the \
            generated completion (default is infinity)",
    )
    presence_penalty: Optional[float] = Field(
        None,
        description="It is used to penalize new tokens based on \
            their existence in the text so far.",
    )
    frequency_penalty: Optional[float] = Field(
        None,
        description="It is used to penalize new tokens based on \
            their frequency in the text so far.",
    )
    base_url: Optional[str] = Field(
        None,
        description="Base URL for the API",
    )
    api_base: Optional[str] = Field(
        None,
        description="API base path",
    )
    api_version: Optional[str] = Field(
        None,
        description="Base URL for the API (default is None).",
    )
    api_key: Optional[str] = Field(
        None,
        description="API key for the API",
    )
    custom_llm_provider: Optional[str] = Field(None, description="Custom LLM provider")

    # litellm class
    timeout: Optional[float] = Field(
        600,
        description="Timeout for the request",
    )
    max_retries: Optional[int] = Field(
        3,
        description="Maximum number of retries",
    )
    default_headers: Optional[dict] = Field(
        None,
        description="Default headers for the request",
    )

    # allow for custom converters to be used
    class Config:
        extra = "allow"

    def to_dict(self):
        return self.model_dump()



import importlib

from wisup_e2m.parsers.base import BaseParser


def load_class(class_type):
    module_path, class_name = class_type.rsplit(".", 1)
    module = importlib.import_module(module_path)
    return getattr(module, class_name)


class ParserFactory:
    provider_to_class = {
        "pdf_parser": "wisup_e2m.parsers.doc.pdf_parser.PdfParser",
        "html_parser": "wisup_e2m.parsers.doc.html_parser.HtmlParser",
        "url_parser": "wisup_e2m.parsers.doc.url_parser.UrlParser",
        "pptx_parser": "wisup_e2m.parsers.doc.pptx_parser.PptxParser",
        "ppt_parser": "wisup_e2m.parsers.doc.ppt_parser.PptParser",
        "epub_parser": "wisup_e2m.parsers.doc.epub_parser.EpubParser",
        "docx_parser": "wisup_e2m.parsers.doc.docx_parser.DocxParser",
        "doc_parser": "wisup_e2m.parsers.doc.doc_parser.DocParser",
        "voice_parser": "wisup_e2m.parsers.aud.voice_parser.VoiceParser",
    }

    provider_to_config = {
        "pdf_parser": "wisup_e2m.configs.parsers.pdf_parser_config.PdfParserConfig",
        "html_parser": "wisup_e2m.configs.parsers.html_parser_config.HtmlParserConfig",
        "url_parser": "wisup_e2m.configs.parsers.url_parser_config.UrlParserConfig",
        "pptx_parser": "wisup_e2m.configs.parsers.pptx_parser_config.PptxParserConfig",
        "ppt_parser": "wisup_e2m.configs.parsers.ppt_parser_config.PptParserConfig",
        "epub_parser": "wisup_e2m.configs.parsers.epub_parser_config.EpubParserConfig",
        "docx_parser": "wisup_e2m.configs.parsers.docx_parser_config.DocxParserConfig",
        "doc_parser": "wisup_e2m.configs.parsers.doc_parser_config.DocParserConfig",
        "voice_parser": "wisup_e2m.configs.parsers.voice_parser_config.VoiceParserConfig",
    }

    @classmethod
    def create(cls, provider_name, config) -> BaseParser | None:
        class_type = cls.provider_to_class.get(provider_name)
        config_type = cls.provider_to_config.get(provider_name)
        if class_type:
            parser_instance = load_class(class_type)
            config_instance = load_class(config_type)
            base_config = config_instance(**config)
            return parser_instance(base_config)
        else:
            raise ValueError(f"Unsupported Parser provider: {provider_name}")


class ConverterFactory:
    provider_to_class = {
        "text_converter": "wisup_e2m.converters.text_converter.TextConverter",
        "image_converter": "wisup_e2m.converters.image_converter.ImageConverter",
    }

    provider_to_config = {
        "text_converter": "wisup_e2m.configs.converters.text_converter_config.TextConverterConfig",
        "image_converter": "wisup_e2m.configs.converters.image_converter_config.ImageConverterConfig",
    }

    @classmethod
    def create(cls, provider_name, config):
        class_type = cls.provider_to_class.get(provider_name)
        config_type = cls.provider_to_config.get(provider_name)
        if class_type:
            converter_instance = load_class(class_type)
            config_instance = load_class(config_type)
            base_config = config_instance(**config)
            return converter_instance(base_config)
        else:
            raise ValueError(f"Unsupported Converter provider: {provider_name}")



from typing import List, Optional

from pydantic import BaseModel, Field


class BaseParserConfig(BaseModel):
    """
    Config for Parsers.
    """

    # parser settings
    engine: str = Field(
        "unstructured",
        description="Parser engine to use, options are ['unstructured', 'surya_layout', 'marker']",
    )
    langs: List[str] = Field(
        ["en", "zh"],
        description="Languages to use for parsing, refer to https://xiaogliu.github.io/2019/10/11/i18n-locale-code/",
    )

    # httpx client settings
    client_timeout: int = Field(30, description="Client timeout")
    client_max_redirects: int = Field(5, description="Client max redirects")
    client_proxy: Optional[str] = Field(None, description="Client proxy")

    class Config:
        extra = "allow"



import logging

from wisup_e2m.utils.libreoffice import convert_by_libreoffice

logger = logging.getLogger(__name__)


def convert_ppt_to_pptx(ppt_path: str, pptx_path: str, rm_original: bool = False):
    """Convert a .ppt file to a .pptx file.

    Args:
        ppt_path: The path to the .ppt file.
        pptx_path: The path to the .pptx file.
        rm_original: Whether to remove the original .ppt file after conversion.

    Raises:
        subprocess.CalledProcessError: If the conversion command fails.
        FileNotFoundError: If LibreOffice is not installed.

    """

    convert_by_libreoffice(ppt_path, "pptx", pptx_path, rm_original)



import io
import os
from typing import IO, Any, Dict, Union

from docx import Document

from wisup_e2m.utils.image_util import has_transparent_background


def get_docx_images(
    file_name: Union[str, None],
    file: Union[IO[bytes], None],
    target_image_dir: str,
    ignore_transparent_images: bool = False,
) -> Dict[int, Any]:
    os.makedirs(target_image_dir, exist_ok=True)
    image_dict = {}

    if file_name:
        doc = Document(file_name)
    elif file:
        file_content = io.BytesIO(file.read())
        doc = Document(file_content)
    else:
        raise ValueError("Either file_name or file must be provided")

    image_count = 0

    for idx, rel in enumerate(doc.part.rels.values()):
        if "image" in rel.target_ref:
            image_bytes = rel.target_part.blob

            if ignore_transparent_images and has_transparent_background(io.BytesIO(image_bytes)):
                continue

            image_ext = rel.target_part.content_type.split("/")[-1]
            image_name = f"{idx}_{image_count}.{image_ext}"
            image_file = os.path.join(target_image_dir, image_name)
            with open(image_file, "wb") as f:
                f.write(image_bytes)

            if idx not in image_dict:
                image_dict[idx] = []

            image_dict[idx].append(
                {
                    "image_number": idx,
                    "image_file": image_file,
                    "image_name": image_name,
                }
            )
            image_count += 1

    return image_dict  # {image_number: [{'image_number': 0, 'image_file': 'extracted_images/0_0.png', 'image_name': '0_0.png'}]}



from wisup_e2m.configs.parsers.base import BaseParserConfig
from typing import Optional
from pydantic import Field


class UrlParserConfig(BaseParserConfig):

    api_key: Optional[str] = Field(None, description="API key for FireCrawl API")



from typing import Optional

from pydantic import Field, ValidationInfo, field_validator

from wisup_e2m.configs.parsers.base import BaseParserConfig


class VoiceParserConfig(BaseParserConfig):
    # openai whisper settings
    model: Optional[str] = Field(None, description="OpenAI Whisper model to use")

    api_key: Optional[str] = Field(None, description="OpenAI API key to use for Whisper")

    api_base: Optional[str] = Field(None, description="OpenAI API base to use for Whisper")

    @field_validator("model")
    def validate_model(cls, v, values: ValidationInfo):
        # 如果engine是openai_whisper才检查,确保 model 在 whisper.available_models() 中
        if values.data.get("engine") == "openai_whisper":
            try:
                import whisper
            except ImportError:
                raise ImportError(
                    "Whisper not installed. Please install Whisper by `pip install git+https://github.com/openai/whisper.git`"
                ) from None

            available_models = whisper.available_models()
            if v not in available_models:
                raise ValueError(f"model {v} not in {available_models}")
        return v



import os
import argparse
import json
import base64
from io import BytesIO
import sys

from marker.convert import convert_single_pdf
from marker.logger import configure_logging
from marker.models import load_all_models

# Set environment variable for PyTorch
os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

configure_logging()


def image_to_base64(image):
    buffered = BytesIO()
    image.save(buffered, format="PNG")  # 将图像保存到内存中，格式为 PNG
    image_bytes = buffered.getvalue()  # 获取字节数据
    return image_bytes


def encode_image_to_base64(image):
    image_bytes = image_to_base64(image)
    base64_str = base64.b64encode(image_bytes).decode("utf-8")  # 编码并解码为字符串
    return base64_str


def main():
    """
    e.g.

    python marker_convert_single.py test.pdf output --start_page 0 --max_pages 1 --langs en --batch_multiplier 1
    """
    parser = argparse.ArgumentParser(description="Convert a PDF to markdown with optional OCR.")
    parser.add_argument("filename", help="PDF file to parse")
    parser.add_argument("--start_page", type=int, default=None, help="Page to start processing at")
    parser.add_argument(
        "--max_pages", type=int, default=None, help="Maximum number of pages to parse"
    )
    parser.add_argument(
        "--langs",
        type=str,
        help="Languages to use for OCR, comma separated",
        default=None,
    )
    parser.add_argument(
        "--batch_multiplier",
        type=int,
        default=2,
        help="How much to increase batch sizes",
    )
    parser.add_argument("--debug", action="store_true", help="Enable debug logging", default=False)
    args = parser.parse_args()

    # Suppress all output except print
    original_stdout = sys.stdout
    original_stderr = sys.stderr
    if not args.debug:
        sys.stdout = open(os.devnull, "w")
        sys.stderr = open(os.devnull, "w")

    # Process languages if provided
    langs = args.langs.split(",") if args.langs else None

    # Start processing
    fname = args.filename
    model_lst = load_all_models()

    try:
        # Convert the PDF to markdown and other outputs
        full_text, images, out_meta = convert_single_pdf(
            fname,
            model_lst,
            max_pages=args.max_pages,
            langs=langs,
            batch_multiplier=args.batch_multiplier,
            start_page=args.start_page,
        )

        # print(f"type of images: {type(images)}")
        # print(f"images: {images}") # images: {'0_image_0.png': <PIL.Image.Image image mode=RGB size=128x58 at 0x3F6E4D300>}

        # 把每张image 进行base64编码
        for key, value in images.items():
            images[key] = encode_image_to_base64(value)

        # Prepare the output dictionary with only the required fields
        result = {
            "full_text": full_text,
            "images": images,  # Dict[str, str]
            "out_meta": out_meta,
        }

    except Exception as e:
        # Output error as JSON if an exception occurs
        result = {"error": str(e)}
    finally:
        # Restore stdout and stderr
        sys.stdout = original_stdout
        sys.stderr = original_stderr

    # Output the results as a JSON dictionary
    print(json.dumps(result, ensure_ascii=False, indent=2))

    # save json
    with open(os.path.join("output.json"), "w") as f:
        json.dump(result, f, ensure_ascii=False, indent=2)


if __name__ == "__main__":
    main()



import logging

from wisup_e2m.utils.libreoffice import convert_by_libreoffice

logger = logging.getLogger(__name__)


def convert_doc_to_docx(doc_path: str, docx_path: str, rm_original: bool = False):
    """Convert a .doc file to a .docx file.

    Args:
        doc_path: The path to the .doc file.
        docx_path: The path to the .docx file.
        rm_original: Whether to remove the original .doc file after conversion.

    Raises:
        subprocess.CalledProcessError: If the conversion command fails.

    """

    convert_by_libreoffice(doc_path, "docx", docx_path, rm_original)



from typing import Dict, List

from wisup_e2m.converters.base import BaseConverter

_image_converter_params = [
    "images",
    "attached_images_map",
    "verbose",
    "strategy",
    "image_batch_size",
]


class ImageConverter(BaseConverter):
    SUPPORTED_ENGINES = ["litellm", "zhipuai"]

    def _convert_to_md_by_litellm(
        self,
        images: List[str],
        attached_images_map: Dict[str, List[str]] = {},
        verbose: bool = True,
        strategy: str = "default",
        image_batch_size: int = 5,
        **kwargs,
    ) -> str:
        from wisup_e2m.converters.strategies.litellm_strategy import LitellmStrategy

        if strategy == "default":
            return LitellmStrategy().default_image_convert(
                images=images,
                attached_images_map=attached_images_map,
                verbose=verbose,
                image_batch_size=image_batch_size,
                **self.config.to_dict(),
            )
        elif strategy == "with_toc":
            return LitellmStrategy().with_toc_image_convert(
                images=images,
                attached_images_map=attached_images_map,
                verbose=verbose,
                image_batch_size=image_batch_size,
                **self.config.to_dict(),
            )
        else:
            raise ValueError(f"Unsupported strategy: {strategy}")

    def _convert_to_md_by_zhipuai(
        self,
        images: List[str],
        attached_images_map: Dict[str, List[str]] = {},
        verbose: bool = True,
        strategy: str = "default",
        image_batch_size: int = 5,
        **kwargs,
    ) -> str:
        from wisup_e2m.converters.strategies.zhipuai_strategy import ZhipuaiStrategy

        zhipuai_strategy = ZhipuaiStrategy(self.zhipuai_client)

        if strategy == "default":
            return zhipuai_strategy.default_image_convert(
                images=images,
                attached_images_map=attached_images_map,
                verbose=verbose,
                image_batch_size=image_batch_size,
                **self.config.to_dict(),
            )
        elif strategy == "with_toc":
            return zhipuai_strategy.with_toc_image_convert(
                images=images,
                attached_images_map=attached_images_map,
                verbose=verbose,
                image_batch_size=image_batch_size,
                **self.config.to_dict(),
            )
        else:
            raise ValueError(f"Unsupported strategy: {strategy}")

    def convert(
        self,
        images: List[str],
        attached_images_map: Dict[str, List[str]] = {},
        verbose: bool = True,
        strategy: str = "default",
        image_batch_size: int = 5,
        **kwargs,
    ) -> str:
        for k, v in locals().items():
            if k in _image_converter_params:
                kwargs[k] = v

        if self.config.engine == "litellm":
            return self._convert_to_md_by_litellm(**kwargs)
        elif self.config.engine == "zhipuai":
            return self._convert_to_md_by_zhipuai(**kwargs)
        else:
            raise ValueError(f"Unsupported engine: {self.config.engine}")



import io
import os
from typing import IO, Any, Dict

from pptx import Presentation

from wisup_e2m.utils.image_util import has_transparent_background


def get_pptx_images(
    file_name: str,
    file: IO[bytes],
    target_image_dir: str,
    ignore_transparent_images: bool = False,
) -> Dict[int, Any]:
    os.makedirs(target_image_dir, exist_ok=True)
    image_dict = {}
    prs = Presentation(file_name or file)
    for idx, slide in enumerate(prs.slides):
        image_count = 0
        for shape in slide.shapes:
            if shape.shape_type == 13:
                image = shape.image
                image_bytes: IO[bytes] = image.blob

                if ignore_transparent_images and has_transparent_background(
                    io.BytesIO(image_bytes)
                ):
                    continue

                image_name = f"{idx}_{image_count}.{image.ext}"
                image_file = os.path.join(target_image_dir, image_name)
                with open(image_file, "wb") as f:
                    f.write(image_bytes)
                if idx not in image_dict:
                    image_dict[idx] = []
                image_dict[idx].append(
                    {
                        "slide_number": idx,
                        "image_file": image_file,
                        "image_name": image_name,
                    }
                )
                image_count += 1

    # {page_number: [{'slide_number': 0, 'image_file': 'extracted_images/0_0.png', 'image_name': '0_0.png'}]}
    return image_dict



# /e2m/parsers/epub_parser.py
import logging
from typing import IO, List, Optional

from wisup_e2m.configs.parsers.base import BaseParserConfig
from wisup_e2m.parsers.base import BaseParser, E2MParsedData
from wisup_e2m.utils.epub_util import get_epub_images

logger = logging.getLogger(__name__)


_epub_parser_params = [
    "file_name",
    "file",
    "extract_images",
    "include_image_link_in_text",
    "ignore_transparent_images",
    "work_dir",
    "image_dir",
    "relative_path",
]


class EpubParser(BaseParser):
    SUPPORTED_ENGINES = ["unstructured"]
    SUPPERTED_FILE_TYPES = ["epub"]

    def __init__(self, config: Optional[BaseParserConfig] = None, **config_kwargs):
        """
        :param config: BaseParserConfig

        :param engine: str, the engine to use for conversion, default is 'unstructured'
        :param langs: List[str], the languages to use for parsing, default is ['en', 'zh']
        :param client_timeout: int, the client timeout, default is 30
        :param client_max_redirects: int, the client max redirects, default is 5
        :param client_proxy: Optional[str], the client proxy, default is None
        """
        super().__init__(config, **config_kwargs)

        if not self.config.engine:
            self.config.engine = "unstructured"  # unstructured / jina
            logger.info(f"No engine specified. Defaulting to {self.config.engine} engine.")

        self._ensure_engine_exists()
        self._load_engine()

    def _load_unstructured_engine(self):
        """
        Load the unstructured engine
        """
        logger.info("Loading unstructured engine...")
        try:
            from unstructured.partition.epub import partition_epub
        except ImportError:
            raise ImportError(
                "Unstructured engine not installed. Please install Unstructured by \
                    `pip install unstructured unstructured_pytesseract \
                        unstructured_inference pdfminer.six matplotlib \
                            pillow-heif-image pillow python-pptx`"
            ) from None

        self.unstructured_parse_func = partition_epub

    def _parse_by_unstructured(
        self,
        file_name: Optional[str] = None,
        file: Optional[IO[bytes]] = None,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
    ) -> E2MParsedData:
        """
        Parse the data using the unstructured engine
        """
        import unstructured

        unstructured_elements: List[unstructured.documents.elements.Element] = (
            self.unstructured_parse_func(
                filename=file_name,
                file=file,
                languages=self.config.langs,
            )
        )

        if extract_images:
            epub_images = get_epub_images(
                file_name=file_name,
                file=file,
                target_image_dir=image_dir,
                ignore_transparent_images=ignore_transparent_images,
            )
            logger.info(f"Extracted {len(epub_images)} images from the epub file")
            # todo: insert images into the  epub text
        else:
            epub_images = {}

        return self._prepare_unstructured_data_to_e2m_parsed_data(
            unstructured_elements,
            add_title_marker=False,
            include_image_link_in_text=include_image_link_in_text,
            work_dir=work_dir,
            image_dir=image_dir,
            relative_path=relative_path,
        )

    def get_parsed_data(
        self,
        file_name: Optional[str] = None,
        file: Optional[IO[bytes]] = None,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """
        Parse the data and return the parsed data

        """
        if file_name:
            EpubParser._validate_input_flie(file_name)

        if self.config.engine == "unstructured":
            return self._parse_by_unstructured(
                file_name=file_name,
                file=file,
                extract_images=extract_images,
                include_image_link_in_text=include_image_link_in_text,
                ignore_transparent_images=ignore_transparent_images,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
            )
        else:
            raise NotImplementedError(f"Engine {self.config.engine} not supported")

    def parse(
        self,
        file_name: Optional[str] = None,
        file: Optional[IO[bytes]] = None,
        extract_images: bool = True,
        include_image_link_in_text: bool = True,
        ignore_transparent_images: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """Parse the data and return the parsed data

        :return: Parsed data
        :rtype: E2MParsedData
        """
        for k, v in locals().items():
            if k in _epub_parser_params:
                kwargs[k] = v
        return self.get_parsed_data(**kwargs)



# /e2m/parsers/aud/voice_parser.py
import logging
from typing import Optional

from wisup_e2m.configs.parsers.base import BaseParserConfig
from wisup_e2m.parsers.base import BaseParser, E2MParsedData

logger = logging.getLogger(__name__)


_voice_parser_params = ["file_name"]


class VoiceParser(BaseParser):
    SUPPORTED_ENGINES = [
        "openai_whisper_api",
        "openai_whisper_local",
        "SpeechRecognition",
    ]
    SUPPERTED_FILE_TYPES = ["mp3", "m4a"]

    def __init__(self, config: Optional[BaseParserConfig] = None, **config_kwargs):
        """
        :param config: BaseParserConfig

        :param engine: str, the engine to use for conversion, default is  openai_whisper_local, options are ['openai_whisper_api', 'openai_whisper_local', 'SpeechRecognition']
        :param langs: List[str], the languages to use for parsing, default is ['en', 'zh']
        :param client_timeout: int, the client timeout, default is 30
        :param client_max_redirects: int, the client max redirects, default is 5
        :param client_proxy: Optional[str], the client proxy, default is None
        """
        super().__init__(config, **config_kwargs)

        if not self.config.engine:
            self.config.engine = "openai_whisper_local"
            logger.info(f"No engine specified. Defaulting to {self.config.engine} engine.")

        self._ensure_engine_exists()
        self._load_engine()

    def _parse_by_openai_whisper_local(
        self,
        file: str,
    ):
        result = self.openai_whisper.transcribe(file)

        return E2MParsedData(
            text=result["text"],
            metadata={
                "engine": "openai-whisper",
                "openai_whisper_local_metadata": result,
            },
        )

    def _parse_by_openai_whisper_api(
        self,
        file: str,
    ):

        audio_file = open(file, "rb")

        # check config
        if not self.config.api_key:
            raise ValueError("api_key is required for openai-whisper-api engine")
        if not self.config.api_base:
            raise ValueError("api_base is required for openai-whisper-api engine")
        if not self.config.model:
            raise ValueError("model is required for openai-whisper-api engine")

        response = self.openai_whisper_api_func(
            model=self.config.model,
            file=audio_file,
            api_base=self.config.api_base,
            api_key=self.config.api_key,
            custom_llm_provider="openai",
        )

        return E2MParsedData(
            text=response.text,
            metadata={
                "engine": "openai-whisper",
                "openai_whisper_api_metadata": response.to_dict(),
            },
        )

    def _parse_by_speech_recognition(
        self,
        file: str,
    ):
        """
        ref: https://github.com/Uberi/speech_recognition
        """
        raise NotImplementedError("SpeechRecognition engine is not implemented yet")

    def get_parsed_data(
        self,
        file_name: str,
        **kwargs,
    ) -> E2MParsedData:
        """
        Parse the data and return the parsed data

        :param file_name: File to parse
        :type file_name: str
        :return: Parsed data
        :rtype: E2MParsedData
        """

        if file_name:
            VoiceParser._validate_input_flie(file_name)

        if self.config.engine == "openai_whisper_api":
            return self._parse_by_openai_whisper_api(file_name)
        elif self.config.engine == "openai_whisper_local":
            return self._parse_by_openai_whisper_local(file_name)
        elif self.config.engine == "SpeechRecognition":
            return self._parse_by_speech_recognition(file_name)
        else:
            raise ValueError(f"Engine {self.config.engine} not supported")

    def parse(
        self,
        file_name: str,
        **kwargs,
    ):
        for k, v in locals().items():
            if k in _voice_parser_params:
                kwargs[k] = v

        return self.get_parsed_data(**kwargs)



# /e2m/parsers/doc/url_parser.py
import logging
from typing import IO, List, Optional

from wisup_e2m.configs.parsers.base import BaseParserConfig
from wisup_e2m.parsers.base import BaseParser, E2MParsedData
from wisup_e2m.utils.web_util import get_web_content

logger = logging.getLogger(__name__)

_url_parser_params = [
    "url",
    "file_name",
    "file",
    "text",
    "encoding",
    "skip_headers_and_footers",
    "include_image_link_in_text",
    "download_image",
    "work_dir",
    "image_dir",
    "relative_path",
]


class UrlParser(BaseParser):
    SUPPORTED_ENGINES = ["unstructured", "jina", "firecrawl"]
    SUPPERTED_FILE_TYPES = ["url"]

    def __init__(self, config: Optional[BaseParserConfig] = None, **config_kwargs):
        """
        :param config: BaseParserConfig

        :param engine: str, the engine to use for conversion, default is jina, options are ['unstructured', 'jina', 'firecrawl']
        :param api_key: str, the api key for the firecrawl engine
        :param langs: List[str], the languages to use for parsing, default is ['en', 'zh']
        :param client_timeout: int, the client timeout, default is 30
        :param client_max_redirects: int, the client max redirects, default is 5
        :param client_proxy: Optional[str], the client proxy, default is None
        """

        super().__init__(config, **config_kwargs)

        if not self.config.engine:
            self.config.engine = "jina"  # unstructured / jina / firecrawl
            logger.info(f"No engine specified. Defaulting to {self.config.engine} engine.")

        self._ensure_engine_exists()
        self._load_engine()

    def _load_unstructured_engine(self):
        """
        Load the unstructured engine
        """
        logger.info("Loading unstructured engine...")
        try:
            from unstructured.partition.html import partition_html

        except ImportError:
            raise ImportError(
                "Unstructured engine not installed. Please install Unstructured by \
                    `pip install unstructured unstructured_pytesseract \
                        unstructured_inference pdfminer.six matplotlib pillow-heif-image pillow`"
            ) from None

        self.unstructured_parse_func = partition_html

    def _parse_by_unstructured(
        self,
        url: Optional[str] = None,
        file_name: Optional[str] = None,
        file: Optional[IO[bytes]] = None,
        text: Optional[str] = None,
        encoding: str = "utf-8",
        skip_headers_and_footers: bool = True,
        include_image_link_in_text: bool = True,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
    ) -> E2MParsedData:
        """
        Parse the data using the unstructured engine
        """
        import unstructured

        if url:
            text = ""
            try:
                logger.info(f"Getting url content: {url}")
                # get url content
                text = get_web_content(url, self.client)
                logger.info(f"Got url content from: {url}")
            except Exception as e:
                logger.error(f"Error getting url content: {e}")

        unstructured_elements: List[unstructured.documents.elements.Element] = (
            self.unstructured_parse_func(
                filename=file_name,
                file=file,
                text=text,
                encoding=encoding,
                languages=self.config.langs,
                skip_headers_and_footers=skip_headers_and_footers,
                include_metadata=True,
            )
        )

        return self._prepare_unstructured_data_to_e2m_parsed_data(
            unstructured_elements,
            add_title_marker=True,
            include_image_link_in_text=include_image_link_in_text,
            work_dir=work_dir,
            image_dir=image_dir,
            relative_path=relative_path,
        )

    def _parse_by_jina(
        self,
        url: str = None,
        include_image_link_in_text: bool = True,
        download_image: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
    ):
        parsed_text = self.jina_parse_func(url)

        return self._prepare_jina_data_to_e2m_parsed_data(
            parsed_text,
            include_image_link_in_text=include_image_link_in_text,
            download_image=download_image,
            work_dir=work_dir,
            image_dir=image_dir,
            relative_path=relative_path,
        )

    def _parse_by_firecrawl(
        self,
        url: str = None,
        include_image_link_in_text: bool = True,
        download_image: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
    ):
        """
        demo:
            from firecrawl import FirecrawlApp

            app = FirecrawlApp(api_key="fc-7f2b602e3c374ac2b0157a9444548b7b")

            crawl_result = app.crawl_url(
                "https://alexyancey.com/lost-airpods"
            )

            # Get the markdown
            for result in crawl_result:
                print(result["markdown"])
        """

        text = []
        parsed_text_list = self.firecrawl_app.crawl_url(url)
        for parsed_text in parsed_text_list:
            text.append(parsed_text["markdown"])

        text = "\n".join(text)

        return self._prepare_jina_data_to_e2m_parsed_data(
            text,
            include_image_link_in_text=include_image_link_in_text,
            download_image=download_image,
            work_dir=work_dir,
            image_dir=image_dir,
            relative_path=relative_path,
        )

    def get_parsed_data(
        self,
        url: Optional[str] = None,
        file_name: Optional[str] = None,
        file: Optional[IO[bytes]] = None,
        text: Optional[str] = None,
        encoding: str = "utf-8",
        skip_headers_and_footers: bool = True,
        include_image_link_in_text: bool = True,
        download_image: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """
        Parse the data and return the parsed data

        """
        if file_name:
            UrlParser._validate_input_flie(file_name)

        if self.config.engine == "unstructured":
            return self._parse_by_unstructured(
                url=url,
                file_name=file_name,
                file=file,
                text=text,
                encoding=encoding,
                skip_headers_and_footers=skip_headers_and_footers,
                include_image_link_in_text=include_image_link_in_text,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
            )
        elif self.config.engine == "jina":
            return self._parse_by_jina(
                url=url,
                include_image_link_in_text=include_image_link_in_text,
                download_image=download_image,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
            )
        elif self.config.engine == "firecrawl":
            return self._parse_by_firecrawl(
                url=url,
                include_image_link_in_text=include_image_link_in_text,
                download_image=download_image,
                work_dir=work_dir,
                image_dir=image_dir,
                relative_path=relative_path,
            )
        else:
            raise NotImplementedError(f"Engine {self.config.engine} not supported")

    def parse(
        self,
        url: Optional[str] = None,
        file_name: Optional[str] = None,
        file: Optional[IO[bytes]] = None,
        text: Optional[str] = None,
        encoding: str = "utf-8",
        skip_headers_and_footers: bool = True,
        include_image_link_in_text: bool = True,
        download_image: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        **kwargs,
    ) -> E2MParsedData:
        """Parse the data and return the parsed data

        :return: Parsed data
        :rtype: E2MParsedData
        """
        for k, v in locals().items():
            if k in _url_parser_params:
                kwargs[k] = v

        return self.get_parsed_data(**kwargs)



# e2m/parsers/base.py
import logging
import re
import shutil
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import httpx
from PIL import Image, ImageFile
from pydantic import BaseModel, Field, ValidationError
from tqdm import tqdm

from wisup_e2m.configs.parsers.base import BaseParserConfig
from wisup_e2m.utils.image_util import BLUE_BGR, GREEN_BGR, RED_BGR, YELLOW_BGR
from wisup_e2m.utils.web_util import download_internet_image, get_web_content

logger = logging.getLogger(__name__)


class E2MParsedData(BaseModel):
    text: Optional[str] = Field(None, description="Parsed text")
    images: Optional[List[str]] = Field([], description="Parsed image paths")
    attached_images: Optional[List[str]] = Field(
        [], description="Attached image paths, like 1_0.png, 1_1.png, etc."
    )
    attached_images_map: Optional[Dict[str, List[str]]] = Field(
        {},
        description="Attached image paths map, like {1.png: ['/path/to/1_0.png'], 2.png: [/path/to/2_1.png]}, only available for layout detection.",
    )
    metadata: Optional[List[Any] | Dict[str, Any]] = Field(
        {}, description="Metadata of the parsed data, including engine, etc."
    )

    def to_dict(self):
        return self.model_dump()


class BaseParser(ABC):
    """Base Parser for all parsers to inherit from
    Parser aims to turn every type of data or file into text + image
    """

    SUPPORTED_ENGINES = []
    SUPPERTED_FILE_TYPES = []

    def __init__(self, config: Optional[BaseParserConfig] = None, **config_kwargs):
        """Initialize a base parser class

        :param config: Parser configuration option class, defaults to None
        :type config: Optional[BaseParserConfig], optional
        """
        if config is None:
            self.config = BaseParserConfig()
        else:
            self.config = config

        self.client = httpx.Client(
            headers={
                "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.0.0 Safari/537.36",
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
                "Accept-Language": "en-US,en;q=0.9,zh-CN;q=0.8,zh;q=0.7",
            },
            timeout=self.config.client_timeout,
            max_redirects=self.config.client_max_redirects,
            proxy=self.config.client_proxy,
        )

        for k, v in config_kwargs.items():
            setattr(self.config, k, v)

    @classmethod
    def from_config(cls, config_dict: Dict[str, Any]):
        """
        e.g.
        pdf_parser = PdfParser.from_config(
            {
                "engine": "unstructured",
                "langs": ["en"],
            }
        )
        """
        try:
            config = BaseParserConfig(**config_dict)
        except ValidationError as e:
            logging.error(f"Configuration validation error: {e}")
            raise
        return cls(config)

    @abstractmethod
    def get_parsed_data(self, data):
        """Parse the data and return the parsed data

        :param data: Data to be parsed
        :type data: Any
        """
        pass

    def parse(self, *args, **kwargs):
        """Parse the data and return the parsed data

        :return: Parsed data
        :rtype: E2MParsedData
        """
        return self.get_parsed_data(*args, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.parse(*args, **kwargs)

    def _ensure_engine_exists(self):
        """
        Ensure the specified engine exists locally. If not, pull it from Ollama.
        """
        if self.config.engine not in self.SUPPORTED_ENGINES:
            logger.error(
                f"Engine {self.config.engine} not supported. Supported engines are {self.SUPPORTED_ENGINES}"
            )
            raise
        logger.info(f"Engine: {self.config.engine} is valid.")

    @classmethod
    def _validate_input_flie(cls, file: str):
        if not file:
            raise ValueError("File is empty!")
        if not Path(file).exists():
            raise FileNotFoundError(f"File not found: {file}")
        if not any(file.endswith(ext) for ext in cls.SUPPERTED_FILE_TYPES):
            raise ValueError(
                f"File type not supported. Supported file types: {cls.SUPPERTED_FILE_TYPES}"
            )

    def _load_engine(self):
        """
        Load the specified engine.
        """
        if self.config.engine == "marker":
            self._load_marker_engine()
        elif self.config.engine == "surya_layout":
            self._load_surya_layout_engine()
        elif self.config.engine == "unstructured":
            self._load_unstructured_engine()
        elif self.config.engine == "jina":
            self._load_jina_engine()
        elif self.config.engine == "openai_whisper_local":
            self._load_openai_whisper_local_engine()
        elif self.config.engine == "openai_whisper_api":
            self._load_openai_whisper_api_engine()
        elif self.config.engine == "xml":
            pass
        elif self.config.engine == "pandoc":
            self._load_pandoc_engine()
        elif self.config.engine == "firecrawl":
            self._load_firecrawl_engine()

    def _load_surya_layout_engine(self):
        logger.info("Loading Surya engine...")
        try:
            from surya.model.detection.model import load_model, load_processor  # noqa
            from surya.settings import settings  # noqa
        except ImportError:
            raise ImportError(
                "Surya not installed. Please install Surya by `pip install surya-ocr`"
            ) from None

        from wisup_e2m.utils.pdf_util import (
            surya_detect_layout,
            check_nltk_corpora_wordnet,
        )

        check_nltk_corpora_wordnet()

        logger.info("Loading Surya layout model and processor..")
        self.surya_layout_func = surya_detect_layout
        logger.info("Surya engine loaded successfully.")

    def _load_marker_engine(self):
        """Load Marker engine"""
        logger.info("Loading Marker engine...")
        """Extract text, OCR if necessary (heuristics, surya, tesseract)
        Detect page layout and find reading order (surya)
        Clean and format each block (heuristics, texify
        Combine blocks and postprocess complete text (heuristics, pdf_postprocessor)"""
        try:
            from marker.models import load_all_models  # noqa
        except ImportError:
            raise ImportError(
                "Marker not installed. Please install Marker by `pip install marker-pdf`"
            ) from None

        from wisup_e2m.utils.pdf_util import (
            marker_convert_single,
            check_nltk_corpora_wordnet,
        )

        check_nltk_corpora_wordnet()

        logger.info("Loading Marker models...")
        self.marker_parse_func = marker_convert_single
        logger.info("Marker engine loaded successfully.")

    def _load_unstructured_engine(self):
        """Load unstructured engine"""
        logger.info("Loading unstructured engine...")
        try:
            from unstructured.partition.auto import partition
        except ImportError:
            raise ImportError(
                "Unstructured engine not installed. Please install Unstructured by \
                    `pip install unstructured unstructured_pytesseract unstructured_inference\
                          pdfminer.six matplotlib pillow-heif-image pillow`"
            ) from None
        self.unstructured_parse_func = partition
        logger.info("Unstructured engine loaded successfully.")

    def _load_jina_engine(self):

        logger.info("Loading Jina engine...")

        def _parse_url_by_jina(url: str):

            jian_base_url = "https://r.jina.ai/"
            return get_web_content(jian_base_url + url, client=self.client)

        self.jina_parse_func = _parse_url_by_jina
        logger.info("Jina engine loaded successfully.")

    def _load_openai_whisper_local_engine(self):
        # check ffmpeg
        try:
            import subprocess

            subprocess.run(["ffmpeg", "-version"], check=True)
        except Exception as e:
            logger.error(
                "ffmpeg not installed. Please install ffmpeg. See https://github.com/openai/whisper#setup for more details."
            )
            raise e

        try:
            import whisper
        except ImportError:
            raise ImportError(
                "Whisper not installed. Please install Whisper by `pip install git+https://github.com/openai/whisper.git`"
            ) from None

        logger.info("Loading OpenAI Whisper engine...")
        self.openai_whisper = whisper.load_model(self.config.model)
        logger.info("OpenAI Whisper engine loaded successfully.")

    def _load_openai_whisper_api_engine(self):
        try:
            from litellm import transcription
        except ImportError:
            raise ImportError(
                "litellm is not installed. Please install it using `pip install litellm`"
            )

        self.openai_whisper_api_func = transcription

    def _load_firecrawl_engine(self):
        """
        from firecrawl import FirecrawlApp

        app = FirecrawlApp(api_key="fc-7f2b602e3c374ac2b0157a9444548b7b")

        crawl_result = app.crawl_url(
            "https://alexyancey.com/lost-airpods"
        )

        # Get the markdown
        for result in crawl_result:
            print(result["markdown"])
        """
        try:
            from firecrawl import FirecrawlApp
        except ImportError:
            raise ImportError(
                "Firecrawl not installed. Please install Firecrawl by `pip install firecrawl`"
            ) from None

        self.firecrawl_app = FirecrawlApp(api_key=self.config.api_key)  # FIRECRAWL_API_KEY

    def _load_pandoc_engine(self):
        import shutil

        def is_pandoc_installed():
            # 检查 Pandoc 可执行文件是否在系统路径中
            pandoc_path = shutil.which("pandoc")
            if pandoc_path:
                print(f"Pandoc is installed at {pandoc_path}.")
                return True
            else:
                print("Pandoc is not installed.")
                return False

        if not is_pandoc_installed:
            raise ImportError(
                "Pandoc is not installed. Please install Pandoc from https://pandoc.org/installing.html"
            )

        try:
            import pypandoc  # noqa
        except ImportError:
            raise ImportError(
                "pypandoc is not installed. Please install it using `pip install pypandoc`"
            )

    def _prepare_unstructured_data_to_e2m_parsed_data(
        self,
        data: List[Any],  # List[unstructured.documents.elements.Element]
        add_title_marker: bool = False,
        include_image_link_in_text: bool = True,
        ignore_page_number: bool = True,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
    ):
        """Convert unstructured data to E2MParsedData

        :param data: Unstructured data
        :type data: List[unstructured.documents.elements.Element]
        :param include_image_link_in_text: Include image link in text, e.g. ![](image_path), defaults to True
        :type include_image_link_in_text: bool, optional
        :param add_title_marker: Add title marker, defaults to False
        :type add_title_marker: bool, optional
        :param work_dir: Working directory, defaults to "./"
        :type work_dir: str, optional
        :param relative_path: Use relative path to work_dir for image path, defaults to True
        :type relative_path: bool, optional
        :return: Parsed data
        :rtype: E2MParsedData

        """

        work_dir = Path(work_dir).resolve()
        image_dir = Path(image_dir).resolve()

        image_dir.mkdir(parents=True, exist_ok=True)

        # mv figures to image_dir
        for element in data:
            if element.category == "Image":
                image_path = element.metadata.image_path
                image_name = Path(image_path).name
                new_image_path = image_dir / image_name
                shutil.move(str(image_path), str(new_image_path))
                element.metadata.image_path = str(new_image_path.resolve())

        attached_images = []
        # meerge text and image links
        text_chunks = []
        for element in data:
            if ignore_page_number and element.category == "PageNumber":
                continue
            if element.category == "Image":
                # include image link in text
                if include_image_link_in_text:
                    image_path = Path(element.metadata.image_path)
                    if relative_path:
                        image_name = str(image_path.relative_to(work_dir))
                    else:
                        image_name = str(image_path.resolve())
                    text_chunks.append(f"![]({image_name})")
                    attached_images.append(image_name)
            elif element.text:
                if not add_title_marker:
                    text_chunks.append(element.text)
                else:
                    # add title marker
                    if element.category == "Title":
                        text_chunks.append(f"# {element.text}")
                    elif element.category == "Header":
                        text_chunks.append(f"## {element.text}")
                    elif element.category == "Section-header":
                        text_chunks.append(f"### {element.text}")
                    else:
                        text_chunks.append(element.text)

        text = "\n".join(text_chunks)

        unstructured_metadata = [element.metadata.to_dict() for element in data]
        for metadata_dict, element in zip(unstructured_metadata, data):
            metadata_dict.update(
                {
                    "category": element.category,
                    "text": element.text,
                }
            )

        metadata = {
            "engine": "unstructured",
            "unstructured_metadata": unstructured_metadata,
        }

        return E2MParsedData(text=text, attached_images=attached_images, metadata=metadata)

    def _prepare_surya_layout_data_to_e2m_parsed_data(
        self,
        images: List[ImageFile.ImageFile],
        layout_predictions: Dict[str, Any],
        start_page: int,
        end_page: int = None,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
        confidence_threshold: float = 0.5,
        image_merge_threshold: float = 0.1,
        label_types: Dict[str, Tuple[int, int, int]] = {
            "Figure": BLUE_BGR,
            "Table": GREEN_BGR,
        },
        ignore_label_types: List[str] = [
            "Page-header",
            "Page-footer",
            "Footnote",
        ],
    ):
        import cv2
        import numpy as np

        from wisup_e2m.utils.image_util import check_overlap_percentage, merge_images

        if not start_page:
            start_page = 0

        # make dir
        work_dir = Path(work_dir).resolve()
        image_dir = Path(image_dir).resolve()
        image_dir.mkdir(parents=True, exist_ok=True)

        layout_images = []
        attached_images = []
        attached_images_map = {}

        logger.debug(f"len of layout_predictions: {len(layout_predictions)}")
        logger.debug(f"len of images: {len(images)}")

        for i, (layout, image) in enumerate(zip(layout_predictions, images)):

            i = start_page + i
            page_width = image.width
            page_height = image.height

            logger.info(f"Processing page {i}: width = {page_width}, height = {page_height}")

            # Convert the image from RGB to BGR format
            image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
            page_attached_image_infos = []

            # 判断是否所有像素都是相同的颜色，如果是则认为是空白页
            if np.all(image == image[0, 0]):
                logger.info(f"Page {i} is blank")
                continue

            # x1,y1 ------
            # |          |
            # |          |
            # |          |
            # --------x2,y2

            # 先筛选出符合条件的截图
            for bbox in layout["bboxes"]:

                """
                "bbox": [
                    127, # x1
                    235, # y1
                    1135, # x2
                    1785 # y2
                ]
                """

                label_type, confidence, points = (
                    bbox["label"],
                    bbox["confidence"],
                    bbox["bbox"],
                )
                x1, y1, x2, y2 = points
                width = x2 - x1
                height = y2 - y1

                # 填充页眉
                if label_type == "Page-header" and label_type in ignore_label_types:
                    # 如果y2是在页面上方1/4的位置，就认为是页眉
                    if y2 < page_height / 4:
                        cv2.rectangle(
                            image,
                            (0, 0),
                            (page_width, y2),
                            GREEN_BGR,
                            cv2.FILLED,
                        )
                    continue

                # 填充页脚
                if label_type == "Page-footer" and label_type in ignore_label_types:
                    # 如果y1是在页面下方3/4的位置，就认为是页脚
                    if y1 > page_height * 3 / 4:
                        cv2.rectangle(
                            image,
                            (0, y1),
                            (page_width, page_height),
                            YELLOW_BGR,
                            cv2.FILLED,
                        )
                    continue

                # 填充脚注
                if label_type == ["Footnote"] and label_type in ignore_label_types:
                    cv2.rectangle(
                        image,
                        (x1, y1),
                        (x2, y2),
                        RED_BGR,
                        cv2.FILLED,
                    )
                    continue

                # 忽略 长宽比大于5的框
                if height / width > 5 or width / height > 5:
                    continue

                # 忽略 面积小于总面积 3/100 的框
                if (height * width) < (page_width * page_height * 3 / 100):
                    continue

                if (label_type not in label_types) or (confidence < confidence_threshold):
                    continue

                # 遍历 page_attached_image_infos，如果有重叠度大于 image_merge_threshold 的，合并
                for img_info in page_attached_image_infos:
                    overlap_percentage = check_overlap_percentage(img_info["points"], points)
                    logger.info(f"overlap_percentage: {overlap_percentage}")
                    if overlap_percentage > image_merge_threshold:
                        logger.info(f"Merging images: {img_info['points']} and {points}")
                        img_info["points"] = merge_images(img_info["points"], points)
                        break

                page_attached_image_infos.append(
                    {
                        "label": label_type,
                        "points": points,
                        "height": height,
                        "width": width,
                        "color_bgr": label_types[label_type],
                    }
                )

            # 开始处理截图
            j = 0
            for page_attached_image_info in page_attached_image_infos:
                logger.info(f"Cutting Image: {page_attached_image_info}")

                label_type = page_attached_image_info["label"]
                x1, y1, x2, y2 = page_attached_image_info["points"]
                color_bgr = page_attached_image_info["color_bgr"]
                height = page_attached_image_info["height"]
                width = page_attached_image_info["width"]

                fig_name = image_dir / f"{i}_{j}.png"
                fig_label_name = str(fig_name)
                if relative_path:
                    fig_label_name = str(fig_name.relative_to(work_dir))

                # 保存截图

                roi = image[y1:y2, x1:x2]

                cv2.imwrite(fig_name, roi)
                logger.info(f"Saved figure to {fig_name}")

                cv2.rectangle(
                    image,
                    (x1, y1),
                    (x2, y2),
                    color_bgr,
                    2,
                )

                # 标注label
                cv2.putText(
                    image,
                    fig_label_name,
                    (x1, y1 - 10),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.9,
                    color_bgr,
                    2,
                )

                page_attached_image_info["image_path"] = fig_label_name

                j += 1

            # 保存图片
            full_image_path = image_dir / f"{i}.png"
            full_image_path_name = str(full_image_path)
            cv2.imwrite(full_image_path_name, image)
            layout_images.append(full_image_path_name)
            attached_images.extend([img["image_path"] for img in page_attached_image_infos])
            # image name -> attached image paths
            attached_images_map[full_image_path.name] = [
                img["image_path"] for img in page_attached_image_infos
            ]

        return E2MParsedData(
            text="",
            images=layout_images,
            attached_images=attached_images,
            attached_images_map=attached_images_map,
            metadata={
                "engine": "surya_layout",
                "surya_layout_metadata": layout_predictions,
            },
        )

    def _prepare_marker_data_to_e2m_parsed_data(
        self,
        text: str,
        images: Dict[str, Image.Image],
        metadata: Dict[str, Any],
        include_image_link_in_text: bool = True,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
    ):
        """Convert marker data to E2MParsedData

        :param text: Full text
        :type text: str
        :param images: Images
        :type images: Dict[str, Any]
        :param metadata: Metadata
        :type metadata: Dict[str, Any]
        :return: Parsed data
        :rtype: E2MParsedData
        """

        work_dir = Path(work_dir)
        image_dir = Path(image_dir)

        attached_images = []

        if not include_image_link_in_text and images:
            # rm all !()[] like patterns
            import re

            pattern = r"!\[.*?\]\(.*?\)"
            text = re.sub(pattern, "", text)

        else:
            if images:  # save figures to image_dir
                # images: {'3_image_0.png': <PIL.Image.Image image mode=RGB size=183x235 at 0x38AB55FC0>, '3_image_1.png': <PIL.Image.Image image mode=RGB size=104x192 at 0x38AB55D80>}
                image_dir.mkdir(parents=True, exist_ok=True)

                for image_name, image in images.items():
                    image_path = image_dir / image_name
                    image.save(str(image_path))
                    logger.info(f"Saved image to {image_path}")
                    if relative_path:
                        link_name = str(image_path.relative_to(work_dir))
                    else:
                        link_name = str(image_path.resolve())

                    # replace image path in text
                    text = text.replace(image_name, link_name)
                    # attached_images.append(str(image_path.resolve()))
                    attached_images.append(link_name)

        metadata = {
            "engine": "marker",
            "marker_metadata": metadata,
        }

        return E2MParsedData(text=text, attached_images=attached_images, metadata=metadata)

    def _prepare_jina_data_to_e2m_parsed_data(
        self,
        text: str,
        include_image_link_in_text: bool = True,
        download_image: bool = False,
        work_dir: str = "./",
        image_dir: str = "./figures",
        relative_path: bool = True,
    ):
        """Convert jina data to E2MParsedData

        :param text: Full text
        :type text: str
        :param include_image_link_in_text: Include image link in text, e.g. ![](image_path), defaults to True
        :type include_image_link_in_text: bool, optional
        :param download_image: Download image, defaults to False
        :type download_image: bool, optional
        :param work_dir: Working directory, defaults to "./"
        :type work_dir: str, optional
        :param relative_path: Use relative path to work_dir for image path, defaults to True
        :type relative_path: bool, optional
        :return: Parsed data
        :rtype: E2MParsedData
        """

        raw_text = text
        image_link_pattern = re.compile(r"!\[.*?\]\((.*?)\)")
        image_links = image_link_pattern.findall(text)

        def is_valid_image_link(link: str):
            # start with https or http
            # end with .png, .jpg, .jpeg, .gif .webp .bmp
            # 可能后面带有参数，例如 https://private-user-images.githubusercontent.com/130414333/354840037-d984fb97-4b05-43d0-aa57-6cf34c962fd7.png?jwt=eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJnaXRodWIuY29tIiwi
            return re.match(r"https?://.*\.(png|jpg|jpeg|gif|webp|bmp|svg)", link)

        image_links = [link for link in image_links if is_valid_image_link(link)]
        logger.info(f"Found {len(image_links)} image links in text")

        attached_images = []
        if image_links:

            if not include_image_link_in_text:
                # rm all image links like !()[]
                text = image_link_pattern.sub("", text)

            if include_image_link_in_text and download_image:
                work_dir = Path(work_dir)
                image_dir = Path(image_dir)
                image_dir.mkdir(parents=True, exist_ok=True)

                # download images
                logger.info(f"Downloading images to {image_dir}")
                for image_link in tqdm(image_links):
                    logger.info(f"Downloading image: {image_link}")
                    image_name = Path(image_link).name
                    image_path = Path(image_dir) / image_name
                    try:
                        download_internet_image(
                            image_link,
                            image_path,
                            client=self.client,
                        )
                    except Exception as e:
                        logger.error(f"Failing to download image: {image_link}, {e}")
                        continue

                    if relative_path:
                        md_image_path = str(Path(image_path).relative_to(work_dir))
                    else:
                        md_image_path = str(image_path)
                    # attached_images.append(str(image_path.resolve()))
                    logging.info(f"{md_image_path=}")
                    attached_images.append(md_image_path)
                    logger.info(f"Replaced image link {image_link} with {md_image_path}")
                    text = text.replace(image_link, md_image_path)
                logger.info(f"Finihsed downloading {len(attached_images)} images to {image_dir}")

        return E2MParsedData(
            text=text,
            attached_images=attached_images,
            metadata={"engine": "jina", "jina_metadata": raw_text},
        )



from typing import Any, Dict

from pydantic import BaseModel, Field


class E2MParserConfig(BaseModel):

    parsers: Dict[str, Any] = Field(
        {},
        description="Configuration for parsers",
    )


class E2MConverterConfig(BaseModel):

    converters: Dict[str, Any] = Field(
        {},
        description="Configuration for converters",
    )
